from curvesim.pool.stableswap.raipool import CurveRaiPool

from .metapool import SimCurveMetaPool


class SimCurveRaiPool(CurveRaiPool, SimCurveMetaPool):
    """Sim interface for Curve RAI metapool"""

    def __init__(self, redemption_prices, *args, **kwargs):
        """
        Parameters
        ----------
        redemption_prices : pandas.DataFrame
            timestamped redemption prices
            (see :meth:`.PoolData.redemption_prices()`)

        Other params as in `CurveRaiPool`.
        """
        self.redemption_prices = redemption_prices
        start_redemption_price = int(redemption_prices.price[0])
        super().__init__(*args, redemption_price=start_redemption_price, **kwargs)

    def prepare_for_trades(self, timestamp):
        """
        Updates the redemption price based on the input timestamp
        before computing and doing trades.

        Parameters
        ----------
        timestamp : datetime.datetime
            the time to sample from
        """
        r = self.redemption_prices.price.asof(timestamp)
        self.rate_multiplier = int(r)



from pandas import DataFrame, concat

from curvesim.metrics.base import PoolMetric
from .pool_parameters import get_pool_parameters
from .pool_state import get_pool_state


class StateLog:
    """
    Logger that records simulation/pool state throughout each simulation run and
    computes metrics at the end of each run.
    """

    __slots__ = [
        "metrics",
        "pool",
        "state_per_run",
        "state_per_trade",
    ]

    def __init__(self, pool, metrics):
        self.pool = pool
        self.metrics = prepare_metrics(metrics, pool)
        self.state_per_run = get_pool_parameters(pool)
        self.state_per_trade = []

    def update(self, **kwargs):
        """Records pool state and any keyword arguments provided."""

        self.state_per_trade.append({"pool_state": get_pool_state(self.pool), **kwargs})

    def get_logs(self):
        """Returns the accumulated log data."""

        df = DataFrame(self.state_per_trade)
        state_per_trade = {col: DataFrame(df[col].to_list()) for col in df}
        return {**self.state_per_run, **state_per_trade}

    def compute_metrics(self):
        """Computes metrics from the accumulated log data."""

        state_logs = self.get_logs()
        metric_data = [metric.compute(state_logs) for metric in self.metrics]
        data_per_trade, summary_data = tuple(zip(*metric_data))  # transpose tuple list

        return (
            DataFrame(self.state_per_run, index=[0]),
            concat(data_per_trade, axis=1),
            concat(summary_data, axis=1),
        )


def prepare_metrics(metrics, pool):
    """
    Applies any neccessary preparations to the input metrics.
    Currently, only updates the pool object for PoolMetrics.
    """
    for metric in metrics:
        if isinstance(metric, PoolMetric):
            metric.set_pool(pool)
    return metrics



"""Miscellaneous utility functions."""
from math import factorial

from numpy import append

from curvesim.logging import get_logger
from curvesim.pool import CurveMetaPool, CurvePool

logger = get_logger(__name__)


def compute_volume_multipliers(pool_vol, market_vol, n, pool_type, mode=1):
    """
    Computes volume multipliers (vol_mult) used for volume limiting.

    Parameters
    ----------
    pool_vol : numpy.ndarray
        Total volume for the pool over the simulation period.

    market_vol : numpy.ndarray
        Total market volume for each token pair over the simulation period.

    n : int
        The number of token-types in the pool (e.g., DAI, USDC, USDT = 3)

    pool_type : str
        "CurvePool" or "CurveMetaPool"

    vol_mode : int, default=1
        Modes for computing the volume multiplier:

        1: limits trade volumes proportionally to market volume for each pair

        2: limits trade volumes equally across pairs

        3: mode 2 for trades with meta-pool asset, mode 1 for basepool-only trades

    """
    if pool_type is CurvePool:
        vol_mult = pool_vol_mult(pool_vol, market_vol, n, mode)

    elif pool_type is CurveMetaPool:
        vol_mult = metapool_vol_mult(pool_vol, market_vol, n, mode)

    else:
        raise TypeError(f"Pool type {pool_type} not supported by this pipeline")

    logger.info(f"Volume Multipliers: {vol_mult}")
    return vol_mult


def pool_vol_mult(pool_vol, market_vol, n, mode):
    if mode == 1:
        vol_mult = pool_vol / market_vol.sum()

    if mode == 2:
        vol_mult = pool_vol.repeat(n) / n / market_vol

    if mode == 3:
        logger.info("Vol_mode=3 only available for meta-pools. Reverting to vol_mode=1")
        vol_mult = pool_vol / market_vol.sum()

    return vol_mult


def metapool_vol_mult(pool_vol, market_vol, n, mode):
    pool_vol_meta = pool_vol[0]
    pool_vol_base = pool_vol[1]
    mkt_vol_meta = market_vol[0 : n[1]]
    mkt_vol_base = market_vol[n[1] :]

    n_base_pairs = int(factorial(n[1]) / 2 * factorial(n[1] - 2))

    if mode == 1:
        vol_mult = append(
            pool_vol_meta / mkt_vol_meta.sum().repeat(n[1]),
            pool_vol_base / mkt_vol_base.sum().repeat(n_base_pairs),
        )

    elif mode == 2:
        vol_mult = append(
            pool_vol_meta.repeat(n[1]) / n[1] / mkt_vol_meta,
            pool_vol_base.repeat(n_base_pairs) / n_base_pairs / mkt_vol_base,
        )

    elif mode == 3:
        vol_mult = append(
            pool_vol_meta.repeat(n[1]) / n[1] / mkt_vol_meta,
            pool_vol_base / mkt_vol_base.sum().repeat(n_base_pairs),
        )

    return vol_mult



"""Common or miscellaneous utility functions"""
import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor

from gmpy2 import mpz


def compute_D(xp, A):
    """Standalone `D` calc neede for some data processing."""
    xp = list(map(int, xp))
    n = len(xp)
    S = sum(xp)
    Dprev = 0
    D = S
    Ann = A * n
    D = mpz(D)
    Ann = mpz(Ann)
    while abs(D - Dprev) > 1:
        D_P = D
        for x in xp:
            D_P = D_P * D // (n * x)
        Dprev = D
        D = (Ann * S + D_P * n) * D // ((Ann - 1) * D + (n + 1) * D_P)

    D = int(D)

    return D


# "extra" event loop for special but important use-cases,
# such as running inside a Jupyter Notebook, which already
# runs an event loop for "convenient" await syntax.
_loop = None


def _setup_extra_event_loop():
    """Sets up the extra event loop for scheduling."""
    # pylint: disable=global-statement
    global _loop
    _loop = asyncio.new_event_loop()
    ThreadPoolExecutor().submit(_loop.run_forever)


def sync(func):
    """
    Returns a sync version of an async function.

    Parameters
    ----------
    func : callable
        An async function.

    Returns
    -------
    inner : callable
        Sync version of the async function.
    """

    @functools.wraps(func)
    def inner(*args, event_loop=None, **kwargs):
        loop = event_loop or asyncio.get_event_loop()
        coro = func(*args, **kwargs)
        if loop.is_running():
            # If for some reason, we are trying to make async code
            # synchronous inside a running event loop, we are
            # probably in something like a Jupyter notebook.
            if not _loop:
                _setup_extra_event_loop()
            future = asyncio.run_coroutine_threadsafe(coro, _loop)
            res = future.result()

        else:
            res = loop.run_until_complete(coro)

        return res

    return inner



from itertools import combinations

from numpy import isnan

from curvesim.utils import override

from ..stableswap import CurvePool
from .simpool import SimStableswapBase


class SimCurvePool(SimStableswapBase, CurvePool):
    @property
    def _precisions(self):
        return self.rates[:]

    @override
    def _init_coin_indices(self):
        return {name: i for i, name in enumerate(self.coin_names)}

    @override
    def price(self, coin_in, coin_out, use_fee=True):
        i, j = self.get_coin_indices(coin_in, coin_out)
        return self.dydx(i, j, use_fee=use_fee)

    @override
    def trade(self, coin_in, coin_out, size):
        i, j = self.get_coin_indices(coin_in, coin_out)

        out_amount, fee = self.exchange(i, j, size)
        # in D units
        volume = size * self._precisions[i] // 10**18

        return out_amount, fee, volume

    @override
    def test_trade(self, coin_in, coin_out, factor, use_fee=True):
        """
        This does the trade but leaves balances unaffected.

        Used to compute liquidity density.
        """
        i, j = self.get_coin_indices(coin_in, coin_out)

        size = self.balances[i] // factor

        with self.use_snapshot_context():
            self.exchange(i, j, size)
            price = self.dydx(i, j, use_fee=use_fee)

        return price

    @override
    def make_error_fns(self):
        # Note: for performance, does not support string coin-names

        xp = self._xp_mem(self.balances, self.rates)

        all_idx = range(self.n_total)
        index_combos = combinations(all_idx, 2)

        def get_trade_bounds(i, j):
            xp_j = int(xp[j] * 0.01)
            high = self.get_y(j, i, xp_j, xp) - xp[i]
            return (0, high)

        def post_trade_price_error(dx, i, j, price_target):
            dx = int(dx) * 10**18 // self.rates[i]

            with self.use_snapshot_context():
                if dx > 0:
                    self.exchange(i, j, dx)

                dydx = self.dydxfee(i, j)

            return dydx - price_target

        def post_trade_price_error_multi(dxs, price_targets, coins):
            with self.use_snapshot_context():
                # Do trades
                for k, pair in enumerate(coins):
                    i, j = pair

                    if isnan(dxs[k]):
                        dx = 0
                    else:
                        dx = int(dxs[k]) * 10**18 // self.rates[i]

                    if dx > 0:
                        self.exchange(i, j, dx)

                # Record price errors
                errors = []
                for k, pair in enumerate(coins):
                    i, j = pair
                    dydx = self.dydxfee(i, j)
                    errors.append(dydx - price_targets[k])

            return errors

        return (
            get_trade_bounds,
            post_trade_price_error,
            post_trade_price_error_multi,
            index_combos,
        )



from altair import Axis, Chart, Color, Scale, condition, selection_point, value
from pandas import DataFrame

from curvesim.exceptions import PlotError
from .chart_properties import PROPERTY_CLASSES


def make_selector(
    field,
    options,
    *,
    labels=None,
    title=None,
    style=None,
    toggle=True,
    sel_idx=None,
):
    title = title or ""
    labels = labels or options

    init_sel = get_initial_selection(field, options, sel_idx)
    selector = selection_point(
        fields=[field], clear=False, value=init_sel, toggle=toggle
    )

    sel_chart = make_selector_chart(field, options, labels, selector, style)

    return sel_chart.properties(title=title), selector


def get_initial_selection(field, options, sel_idx):
    if sel_idx is None:
        init_sel = None

    elif isinstance(sel_idx, int):
        init_sel = [{field: options[sel_idx]}]

    elif sel_idx == "all":
        init_sel = [{field: o} for o in options]

    else:
        raise PlotError(f"sel_idx must be None or int or 'all', not {type(sel_idx)}.")

    return init_sel


def make_selector_chart(field, options, opt_labels, selector, style=None):
    _style = {
        "axis": "y",
        "orient": "right",
        "color": Color(scale=Scale(scheme="viridis"), legend=None),
    }

    if style:
        _style.update(style)

    axis = _style["axis"]
    orient = _style["orient"]
    axis_class = PROPERTY_CLASSES["encoding"][axis]

    encoding = {
        axis: axis_class("labels:O", title=None, axis=Axis(orient=orient)),
        "color": _style["color"],
        "opacity": condition(selector, value(1), value(0.1)),
    }

    df = DataFrame({field: options, "labels": opt_labels})
    return Chart(df).mark_rect().encode(**encoding).add_params(selector)


def if_selected(selected, selector, field, if_true, if_false):
    return condition(
        f"indexof({selector.name}.{field}, '{selected}') != -1", if_true, if_false
    )



"""Unit tests for SimStableswapBase"""
from itertools import combinations

import pytest

from curvesim.pipelines.arbitrage import Arbitrageur
from curvesim.pool.sim_interface.simpool import SimStableswapBase
from curvesim.utils import override

# pylint: disable=redefined-outer-name


class FakeSimStableswap(SimStableswapBase):
    """
    Fake implementation of a subclass of `SimStableswapBase`
    for testing purposes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n = 3

    @override
    def _init_coin_indices(self):
        return {"SYM_0": 0, "SYM_1": 1, "SYM_2": 2}

    @override
    def test_trade(self, coin_in, coin_out, factor):
        pre_price = self.price(coin_in, coin_out)
        # post_price should be less than pre_price
        post_price = 0.90 * pre_price
        return post_price

    @override
    def make_error_fns(self):
        all_idx = range(self.n)
        index_combos = combinations(all_idx, 2)

        # pylint: disable=unused-argument

        def get_trade_bounds(i, j):
            xp = [529818 * 10**18, 760033 * 10**18, 434901 * 10**18]
            high = xp[i]
            return (0, high)

        def post_trade_price_error(dx, i, j, price_target):
            price = self.price(i, j)

            # solver requires opposite signs for the value
            # of the error function on the bounds
            lo, hi = get_trade_bounds(i, j)
            if abs(dx - lo) < 0.000000005:  # pylint: disable=no-else-return
                price = 2
                return price - price_target
            elif abs(dx - hi) < 0.000000005:
                price = 0
                return price - price_target
            else:
                return 0.00000001

        def post_trade_price_error_multi(dxs, price_targets, coins):
            errors = [0.00000001] * len(dxs)
            return errors

        return (
            get_trade_bounds,
            post_trade_price_error,
            post_trade_price_error_multi,
            index_combos,
        )

    @override
    def price(self, coin_in, coin_out, use_fee=True):
        _prices = {
            (0, 1): 1.0002,
            (0, 2): 0.99821,
            (1, 2): 0.989199,
        }
        for (i, j), p in _prices.copy().items():
            _prices[(j, i)] = 1 / p
        return _prices[(coin_in, coin_out)]

    @override
    def trade(self, coin_in, coin_out, size):
        out_amount = size
        fee = 0
        volume = 10**18
        return out_amount, fee, volume


@pytest.fixture(scope="function")
def sim_stableswap():
    """Fixture for testing derived class"""
    return FakeSimStableswap()


def test_sim_stableswap_init(sim_stableswap):
    """Test __init__"""
    assert sim_stableswap.n == 3


def test_sim_stableswap_coin_indices(sim_stableswap):
    """Test index conversion and getting"""
    result = sim_stableswap.get_coin_indices("SYM_2", "SYM_0")
    assert result == [2, 0]

    result = sim_stableswap.get_coin_indices(1, 2)
    assert result == [1, 2]


def test_compute_trades(sim_stableswap):
    """Test error functions with Arbitrageur.compute_trades"""
    trader = Arbitrageur(sim_stableswap)
    prices = [1.00, 1.01, 0.98]
    volume_limits = [100000, 125000, 150000]
    trades, _, _ = trader.compute_trades(prices, volume_limits)
    assert len(trades) == 3
    for t in trades:
        _, _, size = t
        assert size > 0


def test_do_trades(sim_stableswap):
    """Test trade method with Arbitrageur.do_trades"""
    trader = Arbitrageur(sim_stableswap)
    trades = []
    trades_done, volume = trader.do_trades(trades)
    assert trades_done == []  # pylint: disable=use-implicit-booleaness-not-comparison
    assert volume == 0

    trades = [
        (0, 1, 99999999990000015900672),
        (2, 0, 86980634902377172828160),
        (2, 1, 86980634902377172828160),
    ]
    trades_done, volume = trader.do_trades(trades)
    # volume is faked to produce 1 * 10**18 each trade
    assert volume == len(trades) * 10**18
    for t in trades_done:
        # `trade` is faked to produce out amount equal to in amount
        assert t[2] == t[3]



"""
Functions and interfaces used in the simulation pipeline framework.
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from multiprocessing import Pool as cpu_pool
from typing import Union

from numpy import ndarray
from pandas import Series

from curvesim.logging import (
    configure_multiprocess_logging,
    get_logger,
    multiprocessing_logging_queue,
)

logger = get_logger(__name__)


def run_pipeline(param_sampler, price_sampler, strategy, ncpu=4):
    """
    Core function for running pipelines.

    Typically called within a function specifying the pipeline components
    (see, e.g., :func:`.volume_limited_arbitrage()`)

    Parameters
    ----------
    param_sampler : iterator
        An iterator that returns pool parameters (see :mod:`.param_samplers`).

    price_sampler : iterator
        An iterator that returns (minimally) a time-series of prices
        (see :mod:`.price_samplers`).

    strategy: callable
        A function dictating what happens at each timestep.

    ncpu : int, default=4
        Number of cores to use.

    Returns
    -------
    results : tuple
        Contains the metrics produced by the strategy.

    """
    if ncpu > 1:
        price_sampler_data = list(price_sampler)

        with multiprocessing_logging_queue() as logging_queue:
            strategy_args_list = [
                (pool, params, price_sampler_data) for pool, params in param_sampler
            ]

            wrapped_args_list = [
                (strategy, logging_queue, *args) for args in strategy_args_list
            ]

            with cpu_pool(ncpu) as clust:
                results = tuple(
                    zip(*clust.starmap(wrapped_strategy, wrapped_args_list))
                )
                clust.close()
                clust.join()  # coverage needs this

    else:
        results = []
        for pool, params in param_sampler:
            metrics = strategy(pool, params, price_sampler.restart())
            results.append(metrics)
        results = tuple(zip(*results))

    return results


def wrapped_strategy(strategy, logging_queue, *args):
    """
    This wrapper ensures we configure logging to use the
    multiprocessing enqueueing logic within the new process.

    Must be defined at the top-level of the module so it can
    be pickled.
    """
    configure_multiprocess_logging(logging_queue)
    return strategy(*args)


class SimPool(ABC):
    """
    The interface that must be implemented by pools used in simulations.

    See curvesim.pool.sim_interface for implementations.
    """

    def prepare_for_trades(self, timestamp):
        """
        Does any necessary preparation before computing and doing trades.

        The input timestamp can be used to fetch any auxiliary data
        needed to prep the state.

        Base implementation is a no-op.

        Parameters
        ----------
        timestamp : datetime.datetime
            the time to sample from
        """

    @abstractmethod
    def price(self, coin_in, coin_out, use_fee=True):
        """
        Returns the spot price of `coin_in` quoted in terms of `coin_out`,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        The indices are assumed to include base pool underlyer indices.

        Parameters
        ----------
        coin_in : str, int
            ID of coin to be priced; in a swapping context, this is
            the "in"-token.
        coin_out : str, int
            ID of quote currency; in a swapping context, this is the
            "out"-token.
        use_fee: bool, default=False
            Deduct fees.

        Returns
        -------
        float
            Price of `coin_in` quoted in `coin_out`
        """
        raise NotImplementedError

    @abstractmethod
    def trade(self, coin_in, coin_out, size):
        """
        Perform an exchange between two coins.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.
        coin_out : str, int
            ID of "out" coin.
        size : int
            Amount of coin `i` being exchanged.

        Returns
        -------
        (int, int, int)
            (amount of coin `j` received, trading fee, volume)

            Note that coin amounts and fee are in native token units but `volume`
            is normalized to be in the same units as pool value.  This enables
            cross-token comparisons and totaling of volume.
        """
        raise NotImplementedError

    @abstractmethod
    def make_error_fns(self):
        """
        Returns the pricing error functions needed for determining the
        optimal arbitrage in simulations.
        """
        raise NotImplementedError


@dataclass(eq=False, slots=True)
class TradeData:
    trades: list
    volume: int
    volume_limits: Union[list, ndarray, Series]
    price_errors: Union[list, ndarray, Series]



"""
Mainly a module to house the `CryptoPool`, a cryptoswap implementation in Python.
"""
import time
from typing import List

from gmpy2 import mpz

from curvesim.exceptions import CalculationError, CryptoPoolError, CurvesimValueError
from curvesim.pool.base import Pool

NOISE_FEE = 10**5  # 0.1 bps

MIN_GAMMA = 10**10
MAX_GAMMA = 2 * 10**16

EXP_PRECISION = 10**10

PRECISION = 10**18  # The precision to convert to
A_MULTIPLIER = 10000


# pylint: disable-next=too-many-instance-attributes
class CurveCryptoPool(Pool):
    """Cryptoswap implementation in Python."""

    __slots__ = (
        "A",
        "gamma",
        "n",
        "precisions",
        "mid_fee",
        "out_fee",
        "allowed_extra_profit",
        "fee_gamma",
        "adjustment_step",
        "admin_fee",
        "ma_half_time",
        "price_scale",
        "_price_oracle",
        "last_prices",
        "last_prices_timestamp",
        "_block_timestamp",
        "balances",
        "D",
        "virtual_price",
        "tokens",
        "xcp_profit",
        "xcp_profit_a",
        "not_adjusted",
    )

    # pylint: disable-next=too-many-locals,too-many-arguments
    def __init__(
        self,
        A: int,
        gamma: int,
        n: int,
        precisions: List[int],
        mid_fee: int,
        out_fee: int,
        allowed_extra_profit: int,
        fee_gamma: int,
        adjustment_step: int,
        ma_half_time: int,
        price_scale: int,
        price_oracle=None,
        last_prices=None,
        balances=None,
        D=None,
        tokens=None,
        admin_fee: int = 5 * 10**9,
        xcp_profit=10**18,
        xcp_profit_a=10**18,
    ):
        """
        Parameters
        ----------
        A : int
            Amplification coefficient; this is :math:`A n^n` in the whitepaper
            multiplied by 10**4 for greater precision.
        gamma: int
            Decay factor for A.
        n: int
            Number of coins; currently only n = 2 is supported.
        precisions: list of int
            Precision adjustments to convert native token units to 18 decimals;
            this assumes tokens have at most 18 decimals
            i.e. balance in native units * precision = balance in D units
        mid_fee: int
            Fee with 10**10 precision, for trades near price scale
        out_fee: int
            Fee with 10**10 precision, used to adjust `mid_fee` for trades
            further away from price_scale
        allowed_extra_profit: int
            "Buffer" used to determine if the price adjustment algorithm
            should run.
        fee_gamma: int
            Factor used to control the transition from `mid_fee` to `out_fee`.
        adjustment_step:
            Minimum step size to adjust the price scale.
        ma_half_time: int
            "Half-life" for exponential moving average of trade prices.
        price_scale: int
            Price scale value for the pool.  This is where liquidity is concentrated.
        price_oracle: int, optional
            Price oracle value for the pool.  This is the EMA price used to
            adjust the price scale toward.
            Defaults to `price_scale`.
        last_prices: int, optional
            Last trade price for the pool.
            Defaults to `price_scale`.
        balances: list of int, optional
            Coin balances in native token units;
            either `balances` or `D` is required
        D : int, optional
            Stableswap invariant for given balances, precisions, price_scale,
            A, and gamma; either `balances` or `D` is required
        tokens: int, optional
            LP token supply (default is calculated from `D`, which is also
            calculated if needed)
        admin_fee: int, optional
            Percentage of `fee` with 10**10 precision.  Fee paid to the DAO
            (default = 5*10**9)
        xcp_profit: int, optional
            Counter for accumulated profits, no losses (default = 10**18)
        xcp_profit_a: int, optional
            Value of `xcp_profit` when admin fees last claimed (default = 10**18)
        """
        self.A = A
        self.gamma = gamma

        self.mid_fee = mid_fee
        self.out_fee = out_fee
        self.allowed_extra_profit = allowed_extra_profit
        self.fee_gamma = fee_gamma
        self.adjustment_step = adjustment_step
        self.admin_fee = admin_fee

        self.price_scale = price_scale
        self._price_oracle = price_oracle or price_scale
        self.last_prices = last_prices or price_scale
        self.ma_half_time = ma_half_time

        self._block_timestamp = _get_unix_timestamp()
        self.last_prices_timestamp = self._block_timestamp

        self.xcp_profit = xcp_profit
        self.xcp_profit_a = xcp_profit_a  # Full profit at last claim of admin fees
        self.not_adjusted = False

        if n != 2:
            raise CryptoPoolError("Only 2-coin crypto pools are currently supported.")
        self.n = n
        self.precisions = precisions

        if len(precisions) != n:
            raise ValueError("`len(precisions)` must equal `n`")

        if balances is None and D is None:
            raise ValueError("Must provide at least one of `balances` or `D`.")

        if balances:
            self.balances = balances

        if D is not None:
            self.D = D
            if not balances:
                self.balances = [
                    D // n // precisions[0],
                    D * PRECISION // (n * self.price_scale) // precisions[1],
                ]
        else:
            xp = self._xp()
            D = self._newton_D(A, gamma, xp)
            self.D = D

        xcp = self._get_xcp(D)
        if tokens is not None:
            self.tokens = tokens
        else:
            self.tokens = xcp

        self.virtual_price = 10**18 * xcp // tokens

    def _xp(self) -> List[int]:
        """
        Calculate the balances in units of `D`, converting using `price_scale`
        so a unit of each token has equal value.
        """
        precisions = self.precisions
        return [
            self.balances[0] * precisions[0],
            self.balances[1] * precisions[1] * self.price_scale // PRECISION,
        ]

    def _get_xcp(self, D: int) -> int:
        """
        Calculate the constant-product profit, using the balances at
        equilibrium point.
        """
        n_coins: int = self.n
        x: List[int] = [
            D // n_coins,
            D * PRECISION // (self.price_scale * n_coins),
        ]
        return _geometric_mean(x, True)

    # pylint: disable=too-many-locals,too-many-branches
    @staticmethod
    def _newton_D(  # noqa: complexity: 13
        ANN: int,
        gamma: int,
        x_unsorted: List[int],
    ) -> List[int]:
        """
        Finding the `D` invariant using Newton's method.

        ANN is A * N**N from the whitepaper multiplied by the
        factor A_MULTIPLIER.
        """
        n_coins: int = len(x_unsorted)

        # Safety checks
        min_A = n_coins**n_coins * A_MULTIPLIER // 10
        max_A = n_coins**n_coins * A_MULTIPLIER * 100000
        if ANN > max_A or ANN < min_A:
            raise CurvesimValueError("Unsafe value for A")
        if gamma > MAX_GAMMA or gamma < MIN_GAMMA:
            raise CurvesimValueError("Unsafe value for gamma")

        # Initial value of invariant D is that for constant-product invariant
        x: List[int] = x_unsorted
        if x[0] < x[1]:
            x = [x_unsorted[1], x_unsorted[0]]

        assert (
            x[0] > 10**9 - 1 and x[0] < 10**15 * 10**18 + 1
        )  # dev: unsafe values x[0]
        assert x[1] * 10**18 // x[0] > 10**14 - 1  # dev: unsafe values x[i] (input)

        D: int = n_coins * _geometric_mean(x, False)
        S: int = x[0] + x[1]

        D = mpz(D)
        S = mpz(S)
        for _ in range(255):
            D_prev: int = D

            # K0: int = 10**18
            # for _x in x:
            #     K0 = K0 * _x * n_coins / D
            # collapsed for 2 coins
            K0: int = (10**18 * n_coins**2) * x[0] // D * x[1] // D

            _g1k0: int = gamma + 10**18
            if _g1k0 > K0:
                _g1k0 = _g1k0 - K0 + 1
            else:
                _g1k0 = K0 - _g1k0 + 1

            # D / (A * N**N) * _g1k0**2 / gamma**2
            mul1: int = (
                10**18 * D // gamma * _g1k0 // gamma * _g1k0 * A_MULTIPLIER // ANN
            )

            # 2*N*K0 / _g1k0
            mul2: int = (2 * 10**18) * n_coins * K0 // _g1k0

            neg_fprime: int = (
                (S + S * mul2 // 10**18) + mul1 * n_coins // K0 - mul2 * D // 10**18
            )

            # D -= f / fprime
            D_plus: int = D * (neg_fprime + S) // neg_fprime
            D_minus: int = D * D // neg_fprime
            if 10**18 > K0:
                D_minus += D * (mul1 // neg_fprime) // 10**18 * (10**18 - K0) // K0
            else:
                D_minus -= D * (mul1 // neg_fprime) // 10**18 * (K0 - 10**18) // K0

            if D_plus > D_minus:
                D = D_plus - D_minus
            else:
                D = (D_minus - D_plus) // 2

            diff: int = 0
            if D > D_prev:
                diff = D - D_prev
            else:
                diff = D_prev - D
            if diff * 10**14 < max(
                10**16, D
            ):  # Could reduce precision for gas efficiency here
                # Test that we are safe with the next newton_y
                for _x in x:
                    frac: int = _x * 10**18 // D
                    if frac < 10**16 or frac > 10**20:
                        raise CalculationError("Unsafe value for x[i]")
                return int(D)

        raise CalculationError("Did not converge")

    @staticmethod
    def _newton_y(ANN: int, gamma: int, x: List[int], D: int, i: int) -> int:
        """
        Calculating x[i] given other balances x[0..n_coins-1] and invariant D
        ANN = A * N**N
        """
        n_coins: int = len(x)

        # Safety checks
        min_A = n_coins**n_coins * A_MULTIPLIER // 10
        max_A = n_coins**n_coins * A_MULTIPLIER * 100000
        if ANN > max_A or ANN < min_A:
            raise CurvesimValueError("Unsafe value for A")
        if gamma > MAX_GAMMA or gamma < MIN_GAMMA:
            raise CurvesimValueError("Unsafe value for gamma")
        if D > 10**15 * 10**18 or D < 10**17:
            raise CurvesimValueError("Unsafe value for D")

        x_j: int = x[1 - i]
        y: int = D**2 // (x_j * n_coins**2)
        K0_i: int = (10**18 * n_coins) * x_j // D
        # S_i = x_j

        # frac = x_j * 1e18 / D => frac = K0_i / N_COINS
        assert (
            10**16 * n_coins <= K0_i <= 10**20 * n_coins
        )  # dev: unsafe values x[i]

        convergence_limit: int = max(max(x_j // 10**14, D // 10**14), 100)

        y = mpz(y)
        K0_i = mpz(K0_i)
        D = mpz(D)
        for _ in range(255):
            y_prev: int = y

            K0: int = K0_i * y * n_coins // D
            S: int = x_j + y

            _g1k0: int = gamma + 10**18
            if _g1k0 > K0:
                _g1k0 = _g1k0 - K0 + 1
            else:
                _g1k0 = K0 - _g1k0 + 1

            # D / (A * N**N) * _g1k0**2 / gamma**2
            mul1: int = (
                10**18 * D // gamma * _g1k0 // gamma * _g1k0 * A_MULTIPLIER // ANN
            )

            # 2*K0 / _g1k0
            mul2: int = 10**18 + (2 * 10**18) * K0 // _g1k0

            yfprime: int = 10**18 * y + S * mul2 + mul1
            _dyfprime: int = D * mul2
            if yfprime < _dyfprime:
                y = y_prev // 2
                continue

            yfprime -= _dyfprime
            fprime: int = yfprime // y

            # y -= f / f_prime;  y = (y * fprime - f) / fprime
            # y = (yfprime + 10**18 * D - 10**18 * S)
            #   / fprime + mul1 / fprime * (10**18 - K0) / K0
            y_minus: int = mul1 // fprime
            y_plus: int = (yfprime + 10**18 * D) // fprime + y_minus * 10**18 // K0
            y_minus += 10**18 * S // fprime

            if y_plus < y_minus:
                y = y_prev // 2
            else:
                y = y_plus - y_minus

            diff: int = 0
            if y > y_prev:
                diff = y - y_prev
            else:
                diff = y_prev - y
            if diff < max(convergence_limit, y // 10**14):
                frac: int = y * 10**18 // D
                assert 10**16 <= frac <= 10**20  # dev: unsafe value for y
                return int(y)

        raise CalculationError("Did not converge")

    def _increment_timestamp(self, blocks=1, timestamp=None):
        """Update the internal clock used to mimic the block timestamp."""
        if timestamp:
            self._block_timestamp = timestamp
            return

        self._block_timestamp += 12 * blocks

    # pylint: disable=too-many-statements
    def _tweak_price(  # noqa: complexity: 12
        self,
        A: int,
        gamma: int,
        _xp: List[int],
        p_i: int,
        new_D: int,
    ):
        """
        Applies several kinds of updates:
            - EMA price update: price_oracle
            - Profit counters: D, virtual_price, xcp_profit
            - price adjustment: price_scale

        Also claims admin fees if appropriate (enough profit and price scale
        and oracle is close enough).
        """
        price_oracle: int = self._price_oracle
        last_prices: int = self.last_prices
        price_scale: int = self.price_scale
        last_prices_timestamp: int = self.last_prices_timestamp
        block_timestamp: int = self._block_timestamp
        p_new: int = 0
        n_coins: int = self.n

        if last_prices_timestamp < block_timestamp:
            # MA update required
            ma_half_time: int = self.ma_half_time
            alpha: int = _halfpow(
                (block_timestamp - last_prices_timestamp) * 10**18 // ma_half_time
            )
            price_oracle = (
                last_prices * (10**18 - alpha) + price_oracle * alpha
            ) // 10**18
            self._price_oracle = price_oracle
            self.last_prices_timestamp = block_timestamp

        D_unadjusted: int = new_D  # Withdrawal methods know new D already
        if new_D == 0:
            # We will need this a few times (35k gas)
            D_unadjusted = self._newton_D(A, gamma, _xp)

        if p_i > 0:
            last_prices = p_i

        else:
            # calculate real prices
            __xp: List[int] = _xp.copy()
            dx_price: int = __xp[0] // 10**6
            __xp[0] += dx_price
            last_prices = (
                price_scale
                * dx_price
                // (_xp[1] - self._newton_y(A, gamma, __xp, D_unadjusted, 1))
            )

        self.last_prices = last_prices

        total_supply: int = self.tokens
        old_xcp_profit: int = self.xcp_profit
        old_virtual_price: int = self.virtual_price

        # Update profit numbers without price adjustment first
        xp: List[int] = [
            D_unadjusted // n_coins,
            D_unadjusted * PRECISION // (n_coins * price_scale),
        ]
        xcp_profit: int = 10**18
        virtual_price: int = 10**18

        if old_virtual_price > 0:
            xcp: int = _geometric_mean(xp, True)
            virtual_price = 10**18 * xcp // total_supply

            if virtual_price < old_virtual_price:
                raise CryptoPoolError("Loss")

            xcp_profit = old_xcp_profit * virtual_price // old_virtual_price

        self.xcp_profit = xcp_profit

        norm: int = price_oracle * 10**18 // price_scale
        norm = abs(norm - 10**18)
        adjustment_step: int = max(self.adjustment_step, norm // 5)

        needs_adjustment: bool = self.not_adjusted
        # if not needs_adjustment and
        # (virtual_price-10**18 > (xcp_profit-10**18)/2 + self.allowed_extra_profit):
        # (re-arranged for gas efficiency)
        if (
            not needs_adjustment
            and (
                virtual_price * 2 - 10**18
                > xcp_profit + 2 * self.allowed_extra_profit
            )
            and (norm > adjustment_step)
            and (old_virtual_price > 0)
        ):
            needs_adjustment = True
            self.not_adjusted = True

        if needs_adjustment and norm > adjustment_step and old_virtual_price > 0:
            p_new = (
                price_scale * (norm - adjustment_step) + adjustment_step * price_oracle
            ) // norm

            # Calculate balances*prices
            xp = [_xp[0], _xp[1] * p_new // price_scale]

            # Calculate "extended constant product" invariant xCP and virtual price
            D: int = self._newton_D(A, gamma, xp)
            xp = [D // n_coins, D * PRECISION // (n_coins * p_new)]
            new_virtual_price = 10**18 * _geometric_mean(xp, True) // total_supply

            # Proceed if we've got enough profit:
            #   new_virtual_price > 10**18
            #   new_virtual_price - 10**18 > (xcp_profit - 10**18) / 2
            if (new_virtual_price > 10**18) and (
                2 * new_virtual_price - 10**18 > xcp_profit
            ):
                self.price_scale = p_new
                self.D = D
                self.virtual_price = new_virtual_price
                return

        # If we are here, the price_scale adjustment did not happen
        # Still need to update the profit counter and D
        self.D = D_unadjusted
        self.virtual_price = virtual_price

        # norm appeared < adjustment_step after
        if needs_adjustment:
            self.not_adjusted = False
            self._claim_admin_fees()

    def _claim_admin_fees(self):
        # no gulping logic needed for the python code
        xcp_profit: int = self.xcp_profit
        xcp_profit_a: int = self.xcp_profit_a

        vprice: int = self.virtual_price

        if xcp_profit > xcp_profit_a:
            fees: int = (xcp_profit - xcp_profit_a) * self.admin_fee // (2 * 10**10)
            if fees > 0:
                frac: int = vprice * 10**18 // (vprice - fees) - 10**18
                d_supply = self.tokens * frac // 10**18
                self.tokens += d_supply
                xcp_profit -= fees * 2
                self.xcp_profit = xcp_profit

        A = self.A
        gamma = self.gamma
        totalSupply = self.tokens

        D: int = self._newton_D(A, gamma, self._xp())
        self.D = D
        self.virtual_price = 10**18 * self._get_xcp(D) // totalSupply

        if xcp_profit > xcp_profit_a:
            self.xcp_profit_a = xcp_profit

    def get_dy(self, i: int, j: int, dx: int) -> int:
        """
        Calculate the amount received from swapping `dx`
        amount of the `i`-th coin for the `j`-th coin.

        Parameters
        ----------
        i: int
            Index of 'in' coin
        j: int
            Index of 'out' coin
        dx: int
            Amount of 'in' coin
        Returns
        -------
        int
            The 'out' coin amount

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        assert i != j  # dev: same input and output coin
        assert i < self.n  # dev: coin index out of range
        assert j < self.n  # dev: coin index out of range

        xp: List[int] = self.balances.copy()
        xp[i] += dx
        precisions: List[int] = self.precisions
        price_scale: int = self.price_scale * precisions[1]
        xp = [xp[0] * precisions[0], xp[1] * price_scale // PRECISION]

        A = self.A
        gamma = self.gamma
        D: int = self.D

        y: int = self._newton_y(A, gamma, xp, D, j)
        dy: int = xp[j] - y - 1
        xp[j] = y
        if j > 0:
            dy = dy * PRECISION // price_scale
        else:
            dy = dy // precisions[0]
        dy -= self._fee(xp) * dy // 10**10

        return dy

    def _fee(self, xp: List[int]) -> int:
        """
        f = fee_gamma / (fee_gamma + (1 - K))
        where
        K = prod(x) / (sum(x) / N)**N
        (all normalized to 1e18)
        """
        n_coins: int = self.n
        fee_gamma: int = self.fee_gamma
        f: int = xp[0] + xp[1]  # sum
        f = (
            fee_gamma
            * 10**18
            // (
                fee_gamma
                + 10**18
                - (10**18 * n_coins**n_coins) * xp[0] // f * xp[1] // f
            )
        )
        return (self.mid_fee * f + self.out_fee * (10**18 - f)) // 10**18

    def _exchange(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int,
    ) -> int:
        assert i != j  # dev: coin index out of range
        assert i < self.n  # dev: coin index out of range
        assert j < self.n  # dev: coin index out of range
        assert dx > 0  # dev: do not exchange 0 coins

        A = self.A
        gamma = self.gamma
        xp: List[int] = self.balances.copy()
        p: int = 0
        dy: int = 0

        y: int = xp[j]
        x0: int = xp[i]
        xp[i] = x0 + dx
        self.balances[i] = xp[i]

        price_scale: int = self.price_scale
        precisions: List[int] = self.precisions

        xp = [xp[0] * precisions[0], xp[1] * price_scale * precisions[1] // PRECISION]

        prec_i: int = precisions[0]
        prec_j: int = precisions[1]
        if i == 1:
            prec_i = precisions[1]
            prec_j = precisions[0]

        dy = xp[j] - self._newton_y(A, gamma, xp, self.D, j)
        xp[j] -= dy
        dy -= 1

        if j > 0:
            dy = dy * PRECISION // price_scale
        dy = dy // prec_j

        fee = self._fee(xp) * dy // 10**10
        dy -= fee
        assert dy >= min_dy, "Slippage"
        y -= dy

        self.balances[j] = y

        y *= prec_j
        if j > 0:
            y = y * price_scale // PRECISION
        xp[j] = y

        # Calculate price
        if dx > 10**5 and dy > 10**5:
            _dx: int = dx * prec_i
            _dy: int = dy * prec_j
            if i == 0:
                p = _dx * 10**18 // _dy
            else:  # j == 0
                p = _dy * 10**18 // _dx

        self._tweak_price(A, gamma, xp, p, 0)

        return dy, fee

    def exchange(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int = 0,
    ) -> int:
        """
        Swap `dx` amount of the `i`-th coin for the `j`-th coin.

        Parameters
        ----------
        i: int
            'In' coin index
        j: int
            'Out' coin index
        dx: int
            'In' coin amount
        min_dy: int, optional
            Minimum 'out' coin amount required (default = 0)

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)

        Note
        -----
        In the vyper contract, there is an option to exchange using WETH or ETH.
        """
        return self._exchange(
            i,
            j,
            dx,
            min_dy,
        )

    def exchange_underlying(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int = 0,
    ) -> int:
        """
        In the vyper contract, this exchanges using ETH instead of WETH.
        In Curvesim, this is the same as `exchange`.
        """
        return self.exchange(i, j, dx, min_dy)

    def add_liquidity(
        self,
        amounts: List[int],
        min_mint_amount: int = 0,
    ) -> int:
        """
        Add liquidity into the pool by depositing coins for LP tokens.

        Parameters
        ----------
        amounts: List[int]
            Deposit amounts.  At least one coin amount must be nonzero.
        min_mint_amount: int
            Minimum amount of LP tokens required (default = 0)

        Returns
        -------
        int
            Amount of LP tokens minted.
        """
        assert amounts[0] > 0 or amounts[1] > 0  # dev: no coins to add

        A = self.A
        gamma = self.gamma
        n_coins: int = self.n

        xp: List[int] = self.balances.copy()
        amountsp: List[int] = [0] * n_coins
        xx: List[int] = [0] * n_coins
        d_token: int = 0
        d_token_fee: int = 0
        old_D: int = 0

        xp_old: List[int] = xp.copy()

        for i in range(n_coins):
            bal: int = xp[i] + amounts[i]
            xp[i] = bal
            self.balances[i] = bal
        xx = xp.copy()

        precisions: List[int] = self.precisions

        price_scale: int = self.price_scale * precisions[1]
        xp = [xp[0] * precisions[0], xp[1] * price_scale // PRECISION]
        xp_old = [xp_old[0] * precisions[0], xp_old[1] * price_scale // PRECISION]

        for i in range(n_coins):
            if amounts[i] > 0:
                amountsp[i] = xp[i] - xp_old[i]

        old_D = self.D
        D: int = self._newton_D(A, gamma, xp)

        token_supply: int = self.tokens
        if old_D > 0:
            d_token = token_supply * D // old_D - token_supply
        else:
            d_token = self._get_xcp(D)  # making initial virtual price equal to 1
        assert d_token > 0  # dev: nothing minted

        if old_D > 0:
            d_token_fee = self._calc_token_fee(amountsp, xp) * d_token // 10**10 + 1
            d_token -= d_token_fee
            token_supply += d_token
            self.tokens += d_token

            # Calculate price:
            # p_i * (dx_i - dtoken / token_supply * xx_i)
            # = sum{k!=i}(p_k * (dtoken / token_supply * xx_k - dx_k))
            # (simplified for 2 coins)
            p: int = 0
            if d_token > 10**5:
                if amounts[0] == 0 or amounts[1] == 0:
                    S: int = 0
                    precision: int = 0
                    ix: int = 0
                    if amounts[0] == 0:
                        S = xx[0] * precisions[0]
                        precision = precisions[1]
                        ix = 1
                    else:
                        S = xx[1] * precisions[1]
                        precision = precisions[0]
                    S = S * d_token // token_supply
                    p = (
                        S
                        * PRECISION
                        // (
                            amounts[ix] * precision
                            - d_token * xx[ix] * precision // token_supply
                        )
                    )
                    if ix == 0:
                        p = (10**18) ** 2 // p

            self._tweak_price(A, gamma, xp, p, D)

        else:
            self.D = D
            self.virtual_price = 10**18
            self.xcp_profit = 10**18
            self.tokens += d_token

        assert d_token >= min_mint_amount, "Slippage"

        return d_token

    def _calc_token_fee(self, amounts: List[int], xp: List[int]) -> int:
        n_coins: int = self.n
        # fee = sum(amounts_i - avg(amounts)) * fee' / sum(amounts)
        fee: int = self._fee(xp) * n_coins // (4 * (n_coins - 1))
        S: int = 0
        for _x in amounts:
            S += _x
        avg: int = S // n_coins
        Sdiff: int = 0
        for _x in amounts:
            if _x > avg:
                Sdiff += _x - avg
            else:
                Sdiff += avg - _x
        return fee * Sdiff // S + NOISE_FEE

    def remove_liquidity(
        self,
        _amount: int,
        min_amounts=None,
    ):
        """
        Remove liquidity (burn LP tokens) to receive back part (or all) of
        the deposited funds.

        Parameters
        ----------
        _amount: int
            Amount LP tokens to burn.
        min_amounts: List[int], optional
            Minimum required amounts for each coin.  Default is 0 each.

        Note
        ----
        "This withdrawal method is very safe, does no complex math"
        """
        min_amounts = min_amounts or [0, 0]

        total_supply: int = self.tokens
        self.tokens -= _amount
        balances: List[int] = self.balances
        amount: int = _amount - 1  # Make rounding errors favoring other LPs a tiny bit

        for i in range(self.n):
            d_balance: int = balances[i] * amount // total_supply
            assert d_balance >= min_amounts[i]
            self.balances[i] = balances[i] - d_balance

        D: int = self.D
        self.D = D - D * amount // total_supply

    def remove_liquidity_one_coin(
        self, token_amount: int, i: int, min_amount: int
    ) -> int:
        """
        Remove liquidity entirely in one type of coin.
        Fees will be extracted and there may be significant price impact incurred.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.
        min_amount: int
            Minimum amount of the 'out' coin required (default = 0)

        Returns
        -------
        int
            Amount of the `i`-th coin received.
        """
        A = self.A
        gamma = self.gamma

        dy: int = 0
        D: int = 0
        p: int = 0
        xp = [0] * self.n
        dy, p, D, xp = self._calc_withdraw_one_coin(
            A, gamma, token_amount, i, False, True
        )
        assert dy >= min_amount, "Slippage"

        self.balances[i] -= dy
        self.tokens -= token_amount

        self._tweak_price(A, gamma, xp, p, D)

        return dy

    def _calc_withdraw_one_coin(
        self,
        A: int,
        gamma: int,
        token_amount: int,
        i: int,
        update_D: bool,
        calc_price: bool,
    ) -> (int, int, int, List[int]):
        token_supply: int = self.tokens
        assert token_amount <= token_supply  # dev: token amount more than supply
        assert i < self.n  # dev: coin out of range

        xx: List[int] = self.balances.copy()
        D0: int = 0
        precisions: List[int] = self.precisions

        price_scale_i: int = self.price_scale * precisions[1]
        xp: List[int] = [
            xx[0] * precisions[0],
            xx[1] * price_scale_i // PRECISION,
        ]
        if i == 0:
            price_scale_i = PRECISION * precisions[0]

        if update_D:
            D0 = self._newton_D(A, gamma, xp)
        else:
            D0 = self.D

        D: int = D0

        # Charge fee on D, not on y, e.g. reducing invariant LESS than charging user
        fee: int = self._fee(xp)
        dD: int = token_amount * D // token_supply
        D -= dD - (fee * dD // (2 * 10**10) + 1)
        y: int = self._newton_y(A, gamma, xp, D, i)
        dy: int = (xp[i] - y) * PRECISION // price_scale_i
        xp[i] = y

        # Price calc
        p: int = 0
        if calc_price and dy > 10**5 and token_amount > 10**5:
            # p_i = dD / D0 * sum'(p_k * x_k) / (dy - dD / D0 * y0)
            S: int = 0
            precision: int = precisions[0]
            if i == 1:
                S = xx[0] * precisions[0]
                precision = precisions[1]
            else:
                S = xx[1] * precisions[1]
            S = S * dD // D0
            p = S * PRECISION // (dy * precision - dD * xx[i] * precision // D0)
            if i == 0:
                p = (10**18) ** 2 // p

        return dy, p, D, xp

    def calc_withdraw_one_coin(self, token_amount: int, i: int) -> int:
        """
        Calculate the output amount from burning `token amount` of LP tokens
        and receiving entirely in the `i`-th coin.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.

        Returns
        -------
        int
            Output amount of the `i`-th coin.
        """
        return self._calc_withdraw_one_coin(
            self.A, self.gamma, token_amount, i, True, False
        )[0]

    def lp_price(self) -> int:
        """
        Returns an LP token price approximating behavior as a constant-product AMM.
        """
        return (
            2 * self.virtual_price * _sqrt_int(self.internal_price_oracle()) // 10**18
        )

    def internal_price_oracle(self) -> int:
        """
        Return the value of the EMA price oracle.
        """
        price_oracle: int = self._price_oracle
        last_prices_timestamp: int = self.last_prices_timestamp

        block_timestamp: int = self._block_timestamp
        if last_prices_timestamp < block_timestamp:
            ma_half_time: int = self.ma_half_time
            last_prices: int = self.last_prices
            alpha: int = _halfpow(
                (block_timestamp - last_prices_timestamp) * 10**18 // ma_half_time
            )
            return (last_prices * (10**18 - alpha) + price_oracle * alpha) // 10**18

        return price_oracle

    def price_oracle(self) -> int:
        """
        Return the value of the EMA price oracle.

        Same as `internal_price_oracle`.  Kept for compatability with the
        vyper interface.
        """
        return self.internal_price_oracle()

    def get_virtual_price(self) -> int:
        """
        Return the virtual price of an LP token.
        """
        return 10**18 * self._get_xcp(self.D) // self.tokens

    def calc_token_amount(self, amounts: List[int]) -> int:
        """
        Calculate the amount of LP tokens minted by depositing given amounts.

        Parameters
        ----------
        amounts: List[int]
            Deposit amounts.  At least one coin amount must be nonzero.

        Returns
        -------
        int
            Amount of LP tokens minted.
        """
        token_supply: int = self.tokens
        precisions: List[int] = self.precisions
        price_scale: int = self.price_scale * precisions[1]
        A = self.A
        gamma = self.gamma
        xp: List[int] = self._xp()
        amountsp: List[int] = [
            amounts[0] * precisions[0],
            amounts[1] * price_scale // PRECISION,
        ]
        D0: int = self.D
        xp[0] += amountsp[0]
        xp[1] += amountsp[1]
        D: int = self._newton_D(A, gamma, xp)
        d_token: int = token_supply * D // D0 - token_supply
        d_token -= self._calc_token_fee(amountsp, xp) * d_token // 10**10 + 1
        return d_token


def _get_unix_timestamp():
    """Get the timestamp in Unix time."""
    return int(time.time())


def _geometric_mean(unsorted_x: List[int], sort: bool) -> int:
    """
    (x[0] * x[1] * ...) ** (1/N)
    """
    n_coins: int = len(unsorted_x)
    x: List[int] = unsorted_x
    if sort and x[0] < x[1]:
        x = [unsorted_x[1], unsorted_x[0]]

    D: int = mpz(x[0])
    diff: int = 0
    for _ in range(255):
        D_prev: int = D
        # tmp: uint256 = 10**18
        # for _x in x:
        #     tmp = tmp * _x / D
        # D = D * ((N_COINS - 1) * 10**18 + tmp) / (N_COINS * 10**18)
        # line below makes it for 2 coins
        D = (D + x[0] * x[1] // D) // n_coins
        if D > D_prev:
            diff = D - D_prev
        else:
            diff = D_prev - D
        if diff <= 1 or diff * 10**18 < D:
            return int(D)
    raise CalculationError("Did not converge")


def _halfpow(power: int) -> int:
    """
    1e18 * 0.5 ** (power/1e18)

    Inspired by:
    https://github.com/balancer-labs/balancer-core/blob/master/contracts/BNum.sol#L128
    """
    intpow: int = power // 10**18
    otherpow: int = power - intpow * 10**18
    if intpow > 59:
        return 0
    result: int = 10**18 // (2**intpow)
    if otherpow == 0:
        return result

    term: int = mpz(10**18)
    x: int = 5 * 10**17
    S: int = 10**18
    neg: bool = False

    for i in range(1, 256):
        K: int = i * 10**18
        c: int = K - 10**18
        if otherpow > c:
            c = otherpow - c
            neg = not neg
        else:
            c -= otherpow
        term = term * (c * x // 10**18) // K
        if neg:
            S -= term
        else:
            S += term
        if term < EXP_PRECISION:
            return int(result * S // 10**18)

    raise CalculationError("Did not converge")


def _sqrt_int(x: int) -> int:
    """
    Originating from: https://github.com/vyperlang/vyper/issues/1266
    """
    if x == 0:
        return 0

    x = mpz(x)
    z: int = (x + 10**18) // 2
    y: int = x

    for _ in range(256):
        if z == y:
            return int(y)
        y = z
        z = (x * 10**18 // z + z) // 2

    raise CalculationError("Did not converge")



from copy import deepcopy
from functools import partial

from altair import Color, MarkDef, X, Y


def make_chart_kwargs(default, override_dicts):
    chart_kwargs = deepcopy(default)

    for d in override_dicts:
        for key, val in d.items():
            update = UPDATE_RULES.get(key, update_properties)
            update(chart_kwargs, key, val)

    return init_properties(chart_kwargs)


def update_properties(prop_dict, key, val, depth=0, count=0):
    if depth in [0, count]:
        prop_dict[key] = val
    else:
        prop_dict.setdefault(key, {})
        for k, v in val.items():
            update_properties(prop_dict[key], k, v, depth, count + 1)


def concat_properties(prop_dict, key, val):
    prop_dict.setdefault(key, [])

    if isinstance(val, list):
        prop_dict[key] += val
    else:
        prop_dict[key].append(val)


def ignore_property(prop_dict, key, val):  # pylint: disable=unused-argument
    pass


def init_properties(prop_dict, classes=None):
    classes = classes or PROPERTY_CLASSES
    for key, val in prop_dict.items():
        try:
            prop_class = classes[key]
        except KeyError:
            continue

        if isinstance(prop_class, dict):
            prop_dict[key] = init_properties(val, classes=classes[key])
        else:
            prop_dict[key] = prop_class(**val)

    return prop_dict


PROPERTY_CLASSES = {
    "encoding": {
        "x": X,
        "y": Y,
        "color": Color,
    },
    "mark": MarkDef,
}

UPDATE_RULES = {
    "mark": partial(update_properties, depth=1),
    "encoding": partial(update_properties, depth=2),
    "selection": update_properties,
    "transform": concat_properties,
    "resample": ignore_property,
    "style": ignore_property,
}



"""
Mainly a module to house the `Pool`, a basic stableswap implementation in Python.
"""
from math import prod

from gmpy2 import mpz

from curvesim.pool.snapshot import CurvePoolBalanceSnapshot

from ..base import Pool


class CurvePool(Pool):  # pylint: disable=too-many-instance-attributes
    """
    Basic stableswap implementation in Python.
    """

    snapshot_class = CurvePoolBalanceSnapshot

    __slots__ = (
        "A",
        "n",
        "fee",
        "rates",
        "balances",
        "tokens",
        "fee_mul",
        "admin_fee",
        "r",
        "n_total",
        "admin_balances",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        A,
        D,
        n,
        rates=None,
        tokens=None,
        fee=4 * 10**6,
        fee_mul=None,
        admin_fee=0 * 10**9,
    ):
        """
        Parameters
        ----------
        A : int
            Amplification coefficient; this is :math:`A n^{n-1}` in the whitepaper.
        D : int or list of int
            virtual total balance or pool coin balances in native token units
        n: int
            number of coins
        rates: list of int
            precision and rate adjustments
        tokens: int
            LP token supply
        fee: int, optional
            fee with 10**10 precision (default = .004%)
        fee_mul: optional
            fee multiplier for dynamic fee pools
        admin_fee: int, optional
            percentage of `fee` with 10**10 precision (default = 50%)
        """
        # FIXME: set admin_fee default back to 5 * 10**9
        # once sim code is updated.  Right now we use 0
        # to pass the CI tests.
        rates = rates or [10**18] * n

        if isinstance(D, list):
            balances = D
        else:
            balances = [D // n * 10**18 // _p for _p in rates]

        self.A = A
        self.n = n
        self.fee = fee
        self.rates = rates
        self.balances = balances
        self.tokens = tokens or self.D()
        self.fee_mul = fee_mul
        self.admin_fee = admin_fee
        self.r = False
        self.n_total = n
        self.admin_balances = [0] * n

    def _xp(self):
        rates = self.rates
        balances = self.balances
        return self._xp_mem(rates, balances)

    def _xp_mem(self, rates, balances):
        return [x * p // 10**18 for x, p in zip(balances, rates)]

    def D(self, xp=None):
        """
        `D` is the stableswap invariant; this can be thought of as the value of
        all coin balances if the pool were to become balanced.

        Convenience wrapper for `get_D` which uses the set `A` and makes `xp`
        an optional arg.

        Parameters
        ----------
        xp: list of ints
            Coin balances in units of D

        Returns
        -------
        int
            The stableswap invariant, `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A = self.A
        xp = xp or self._xp()
        return self.get_D(xp, A)

    def get_D(self, xp, A):
        r"""
        Calculate D invariant iteratively using non-overflowing integer operations.

        Stableswap equation:

        .. math::
             A n^n \sum{x_i} + D = A n^n D + D^{n+1} / (n^n \prod{x_i})

        Converging solution using Newton's method:

        .. math::
             d_{j+1} = (A n^n \sum{x_i} + n d_j^{n+1} / (n^n \prod{x_i}))
                     / (A n^n + (n+1) d_j^n/(n^n \prod{x_i}) - 1)

        Replace :math:`A n^n` by `An` and :math:`d_j^{n+1}/(n^n \prod{x_i})` by
        :math:`D_p` to arrive at the iterative formula in the code.

        Parameters
        ----------
        xp: list of ints
            Coin balances in units of D
        A: int
            Amplification coefficient

        Returns
        -------
        int
            The stableswap invariant, `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """  # noqa
        Dprev = 0
        S = sum(xp)
        D = S
        n = self.n
        Ann = A * n
        D = mpz(D)
        Ann = mpz(Ann)
        while abs(D - Dprev) > 1:
            D_P = D
            for x in xp:
                D_P = D_P * D // (n * x)
            Dprev = D
            D = (Ann * S + D_P * n) * D // ((Ann - 1) * D + (n + 1) * D_P)

        D = int(D)
        return D

    def get_D_mem(self, balances, A):
        """
        Convenience wrapper for `get_D` which takes in balances in token units.
        Naming is based on the vyper equivalent.

        Parameters
        ----------
        balances: list of ints
            Coin balances in native token units
        A: int
            Amplification coefficient

        Returns
        -------
        int
            The stableswap invariant, `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        xp = [x * p // 10**18 for x, p in zip(balances, self.rates)]
        return self.get_D(xp, A)

    def get_y(self, i, j, x, xp):
        r"""
        Calculate x[j] if one makes x[i] = x.

        The stableswap equation gives the following:

        .. math::
            x_1^2 + x_1 (\operatorname{sum'} - (A n^n - 1) D / (A n^n))
               = D^{n+1}/(n^{2 n} \operatorname{prod'} A)

        where :math:`\operatorname{sum'}` is the sum of all :math:`x_i` for
        :math:`i \\neq j` and :math:`\operatorname{prod'}` is the product
        of all :math:`x_i` for :math:`i \\neq j`.

        This is a quadratic equation in :math:`x_j`.

        .. math:: x_1^2 + b x_1 = c

        which can then be solved iteratively by Newton's method:

        .. math:: x_1 := (x_1^2 + c) / (2 x_1 + b)

        Parameters
        ----------
        i: int
            index of coin; usually the "in"-token
        j: int
            index of coin; usually the "out"-token
        x: int
            balance of i-th coin in units of D
        xp: list of int
            coin balances in units of D

        Returns
        -------
        int
            The balance of the j-th coin, in units of D, for the other
            coin balances given.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """  # noqa
        xx = xp[:]
        D = self.D(xx)
        D = mpz(D)
        xx[i] = x  # x is quantity of underlying asset brought to 1e18 precision
        n = self.n
        xx = [xx[k] for k in range(n) if k != j]
        Ann = self.A * n
        c = D
        for y in xx:
            c = c * D // (y * n)
        c = c * D // (n * Ann)
        b = sum(xx) + D // Ann - D
        y_prev = 0
        y = D
        while abs(y - y_prev) > 1:
            y_prev = y
            y = (y**2 + c) // (2 * y + b)
        y = int(y)
        return y  # result is in units for D

    def get_y_D(self, A, i, xp, D):
        """
        Calculate x[i] if one uses a reduced `D` than one calculated for given `xp`.

        See docstring for `get_y`.

        Parameters
        ----------
        A: int
            Amplification coefficient for given xp and D
        i: int
            index of coin to calculate balance for
        xp: list of int
            coin balances in units of D
        D: int
            new invariant value

        Returns
        -------
        int
            The balance of the i-th coin, in units of D

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        D = mpz(D)
        n = self.n
        xx = [xp[k] for k in range(n) if k != i]
        S = sum(xx)
        Ann = A * n
        c = D
        for y in xx:
            c = c * D // (y * n)
        c = c * D // (n * Ann)
        b = S + D // Ann
        y_prev = 0
        y = D
        while abs(y - y_prev) > 1:
            y_prev = y
            y = (y**2 + c) // (2 * y + b - D)
        y = int(y)
        return y  # result is in units for D

    def exchange(self, i, j, dx):
        """
        Perform an exchange between two coins.
        Index values can be found via the `coins` public getter method.

        Parameters
        ----------
        i : int
            Index of "in" coin.
        j : int
            Index of "out" coin.
        dx : int
            Amount of coin `i` being exchanged.

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)


        Examples
        --------

        >>> pool = Pool(A=250, D=1000000 * 10**18, n=2, p=[10**30, 10**30])
        >>> pool.exchange(0, 1, 150 * 10**6)
        (149939820, 59999)
        """
        xp = self._xp()
        x = xp[i] + dx * self.rates[i] // 10**18
        y = self.get_y(i, j, x, xp)
        dy = xp[j] - y - 1

        if self.fee_mul is None:
            fee = dy * self.fee // 10**10
        else:
            fee = dy * self.dynamic_fee((xp[i] + x) // 2, (xp[j] + y) // 2) // 10**10

        admin_fee = fee * self.admin_fee // 10**10

        # Convert all to real units
        rate = self.rates[j]
        dy = (dy - fee) * 10**18 // rate
        fee = fee * 10**18 // rate
        admin_fee = admin_fee * 10**18 // rate
        assert dy >= 0

        self.balances[i] += dx
        self.balances[j] -= dy + admin_fee
        self.admin_balances[j] += admin_fee
        return dy, fee

    def calc_withdraw_one_coin(self, token_amount, i, use_fee=True):
        """
        Calculate the amount in the i-th coin received from
        redeeming the given amount of LP tokens.

        By default, fees are deducted.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to redeem
        i: int
            Index of coin to withdraw in.
        use_fee: bool, default=True
            Deduct fees.

        Returns
        -------
        int
            Redemption amount in i-th coin

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A = self.A
        xp = self._xp()
        D0 = self.D()
        D1 = D0 - token_amount * D0 // self.tokens

        new_y = self.get_y_D(A, i, xp, D1)
        dy_before_fee = (xp[i] - new_y) * 10**18 // self.rates[i]

        xp_reduced = xp
        if self.fee and use_fee:
            n_coins = self.n
            _fee = self.fee * n_coins // (4 * (n_coins - 1))

            for j in range(n_coins):
                dx_expected = 0
                if j == i:
                    dx_expected = xp[j] * D1 // D0 - new_y
                else:
                    dx_expected = xp[j] - xp[j] * D1 // D0
                xp_reduced[j] -= _fee * dx_expected // 10**10

        dy = xp[i] - self.get_y_D(A, i, xp_reduced, D1)
        dy = (dy - 1) * 10**18 // self.rates[i]

        if use_fee:
            dy_fee = dy_before_fee - dy
            return dy, dy_fee

        return dy

    def add_liquidity(self, amounts):
        """
        Deposit coin amounts for LP token.

        Parameters
        ----------
        amounts: list of int
            Coin amounts to deposit

        Returns
        -------
        int
            LP token amount received for the deposit amounts.
        """
        mint_amount, fees = self.calc_token_amount(amounts, use_fee=True)
        self.tokens += mint_amount

        balances = self.balances
        afee = self.admin_fee
        admin_fees = [f * afee // 10**10 for f in fees]
        new_balances = [
            bal + amt - fee for bal, amt, fee in zip(balances, amounts, admin_fees)
        ]
        self.balances = new_balances
        self.admin_balances = [t + a for t, a in zip(self.admin_balances, admin_fees)]

        return mint_amount

    def remove_liquidity_one_coin(self, token_amount, i):
        """
        Redeem given LP token amount for the i-th coin.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to redeem
        i: int
            Index of coin to withdraw in

        Returns
        -------
        int
            Redemption amount in i-th coin
        """
        dy, dy_fee = self.calc_withdraw_one_coin(token_amount, i, use_fee=True)
        admin_fee = dy_fee * self.admin_fee // 10**10
        self.balances[i] -= dy + admin_fee
        self.admin_balances[i] += admin_fee
        self.tokens -= token_amount
        return dy, dy_fee

    def calc_token_amount(self, amounts, use_fee=False):
        """
        Calculate the amount of LP tokens received for the given coin
        deposit amounts.

        Fee logic is based on add_liquidity, which makes this more accurate than
        the `calc_token_amount` in the actual contract, which neglects fees.

        By default, it's assumed you the same behavior as the vyper contract,
        which is to NOT deduct fees.

        Parameters
        ----------
        amounts: list of int
            Coin amounts to be deposited.
        use_fee: bool, default=False
            Deduct fees.

        Returns
        -------
        int
            LP token amount received for the deposit amounts.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A = self.A
        old_balances = self.balances
        D0 = self.get_D_mem(old_balances, A)

        new_balances = self.balances[:]
        for i in range(self.n):
            new_balances[i] += amounts[i]
        D1 = self.get_D_mem(new_balances, A)

        mint_balances = new_balances[:]

        if use_fee:
            _fee = self.fee * self.n // (4 * (self.n - 1))

            fees = [0] * self.n
            for i in range(self.n):
                ideal_balance = D1 * old_balances[i] // D0
                difference = abs(ideal_balance - new_balances[i])
                fees[i] = _fee * difference // 10**10
                mint_balances[i] -= fees[i]

        D2 = self.get_D_mem(mint_balances, A)

        mint_amount = self.tokens * (D2 - D0) // D0

        if use_fee:
            return mint_amount, fees

        return mint_amount

    def get_virtual_price(self):
        """
        Returns the expected value of one LP token if the pool were
        to become perfectly balanced (all coins revert to peg).

        Returns
        -------
        int
            Amount of the stableswap invariant, `D`, corresponding to
            one LP token, in units of `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self.D() * 10**18 // self.tokens

    def dynamic_fee(self, xpi, xpj):
        xps2 = xpi + xpj
        xps2 *= xps2  # Doing just ** 2 can overflow apparently
        return (self.fee_mul * self.fee) // (
            (self.fee_mul - 10**10) * 4 * xpi * xpj // xps2 + 10**10
        )

    def dydxfee(self, i, j):
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Trading fees are deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin with fees deducted.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self.dydx(i, j, use_fee=True)

    def dydx(self, i, j, use_fee=False):
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Defaults to no fees deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        xp = self._xp()
        return self._dydx(i, j, xp, use_fee)

    def _dydx(self, i, j, xp, use_fee):
        xi = xp[i]
        xj = xp[j]
        n = self.n
        A = self.A
        D = self.D(xp)
        D_pow = mpz(D) ** (n + 1)
        x_prod = prod(xp)
        A_pow = A * n ** (n + 1)
        dydx = (xj * (xi * A_pow * x_prod + D_pow)) / (
            xi * (xj * A_pow * x_prod + D_pow)
        )

        if use_fee:
            if self.fee_mul is None:
                fee_factor = self.fee / 10**10
            else:
                fee_factor = self.dynamic_fee(xi, xj) / 10**10
        else:
            fee_factor = 0

        dydx *= 1 - fee_factor

        return float(dydx)



"""Pool fixtures to test against vyper implementation.  Uses titanoboa."""
# pylint: disable=redefined-outer-name
import os

import boa
import pytest

_base_dir = os.path.dirname(__file__)
_curve_dir = os.path.join(_base_dir, "curve")
FAKE_ADDRESS = "0xCAFECAFECAFECAFECAFECAFECAFECAFECAFECAFE"


@pytest.fixture(scope="session")
def mainnet_3pool_state():
    """Snapshot of Mainnet 3Pool values"""
    p = [10**18, 10**30, 10**30]
    balances = [
        295949605740077243186725223,
        284320067518878,
        288200854907854,
    ]
    virtual_balances = [b * p // 10**18 for b, p in zip(balances, p)]

    return {
        "N_COINS": 3,
        "A": 2000,
        "p": p,
        "balances": balances,
        "virtual_balances": virtual_balances,
        "lp_tokens": 849743149250065202008212976,
        "virtual_price": 1022038799187029697,
    }


@pytest.fixture(scope="session")
def _vyper_3pool(mainnet_3pool_state):
    """Initialize vyper fixture using mainnet values."""
    lp_total_supply = mainnet_3pool_state["lp_tokens"]
    mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    lp_token = boa.load(mock_filepath, lp_total_supply)

    pool_filepath = os.path.join(_curve_dir, "basepool.vy")
    owner = FAKE_ADDRESS
    coins = [FAKE_ADDRESS] * 3
    A = mainnet_3pool_state["A"]
    fee = 4 * 10**6
    admin_fee = 5 * 10**9
    pool = boa.load(pool_filepath, owner, coins, lp_token, A, fee, admin_fee)

    balances = mainnet_3pool_state["balances"]
    pool.eval(f"self.balances={balances}")

    return pool


@pytest.fixture(scope="session")
def _vyper_metapool(_vyper_3pool):
    """Initialize vyper fixture using mainnet values."""
    metapool_filepath = os.path.join(_curve_dir, "metapool.vy")
    name = "SIM-3Pool"
    symbol = "SIM3CRV-f"
    coin = FAKE_ADDRESS
    rate_multiplier = 10**34  # 2 decimals
    basepool = _vyper_3pool.address
    basepool_token = _vyper_3pool.token()
    A = 1000
    fee = 4 * 10**6
    # Admin fee is hard-coded as 50% for factory pools
    # admin_fee = 5 * 10**9
    metapool = boa.load(
        metapool_filepath,
        name,
        symbol,
        coin,
        basepool,
        basepool_token,
        rate_multiplier,
        A,
        fee,
    )

    balances = [762951074, 12971664836474542835562756]
    metapool.eval(f"self.balances={balances}")
    total_supply = 20312687702458911532611097
    metapool.eval(f"self.totalSupply={total_supply}")

    return metapool


@pytest.fixture(scope="session")
def _cryptopool_lp_token():
    lp_total_supply = 16060447504694332256465310
    mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    lp_token = boa.load(mock_filepath, lp_total_supply)
    return lp_token


@pytest.fixture(scope="session")
def _vyper_cryptopool(_cryptopool_lp_token):
    """
    Initialize vyper fixture for crypto pool
    using default volatile pair settings
    """
    cryptopool_filepath = os.path.join(_curve_dir, "cryptopool.vy")
    coins = [FAKE_ADDRESS] * 2

    # settings based on STG/USDC pool
    # https://etherscan.io/address/0x3211c6cbef1429da3d0d58494938299c92ad5860

    # lp_total_supply = 16060447504694332256465310
    # mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    # lp_token = boa.load(mock_filepath, lp_total_supply)

    A = 400000
    gamma = 72500000000000
    # unpacked_precisions = [10**0, 10**12]
    precisions = 12 << 8
    precisions = precisions | 0  # for explicitness
    mid_fee = 26000000
    out_fee = 45000000
    allowed_extra_profit = 2000000000000
    fee_gamma = 230000000000000
    adjustment_step = 146000000000000
    admin_fee = 5000000000
    ma_half_time = 600
    initial_price = 1550997347493624157

    cryptopool = boa.load(
        cryptopool_filepath,
        A,
        gamma,
        mid_fee,
        out_fee,
        allowed_extra_profit,
        fee_gamma,
        adjustment_step,
        admin_fee,
        ma_half_time,
        initial_price,
        _cryptopool_lp_token,
        coins,
        precisions,
    )

    balances = [20477317313816545807568241, 13270936465339]
    cryptopool.eval(f"self.balances={balances}")
    D = 41060496962103963853877954
    virtual_price = 1026434015737186294
    cryptopool.eval(f"self.D={D}")
    cryptopool.eval(f"self.virtual_price={virtual_price}")
    xcp_profit = 1052829794354693246
    xcp_profit_a = 1052785575319598710
    cryptopool.eval(f"self.xcp_profit={xcp_profit}")
    cryptopool.eval(f"self.xcp_profit_a={xcp_profit_a}")

    return cryptopool


@pytest.fixture(scope="function")
def vyper_3pool(_vyper_3pool):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_3pool


@pytest.fixture(scope="function")
def vyper_metapool(_vyper_metapool):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_metapool


@pytest.fixture(scope="function")
def cryptopool_lp_token(_cryptopool_lp_token):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _cryptopool_lp_token


@pytest.fixture(scope="function")
def vyper_cryptopool(_vyper_cryptopool):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_cryptopool



"""Base SimPool implementation for Curve stableswap pools, both regular and meta."""
from abc import abstractmethod

from curvesim.exceptions import CurvesimValueError
from curvesim.pipelines.templates import SimPool
from curvesim.utils import cache


class SimStableswapBase(SimPool):
    """
    This base class contains common logic useful for all Curve
    stableswap implementations used in arbitrage pipelines:

    - translate from coin names to Curve pool indices
    - compute liquidity density of a coin pair and price-depth
    """

    @abstractmethod
    def _init_coin_indices(self):
        """Produces the coin ID to index mapping from the pool's metadata."""
        raise NotImplementedError

    @property
    @cache
    def coin_indices(self):
        """Return dict mapping coin ID to index."""
        return self._init_coin_indices()

    def get_coin_indices(self, *coins_ids):
        """
        Gets the pool indices for the input coin names.
        Uses the coin_indices set by _init_coin_indices.
        """
        coin_indices = []
        for coin_id in coins_ids:
            if isinstance(coin_id, str):
                coin_id = self.coin_indices[coin_id]
            coin_indices.append(coin_id)

        if len(coin_indices) != len(set(coin_indices)):
            raise CurvesimValueError("Duplicate coin indices.")

        return coin_indices

    @abstractmethod
    def price(self, coin_in, coin_out, use_fee=True):
        raise NotImplementedError

    @abstractmethod
    def trade(self, coin_in, coin_out, size):
        raise NotImplementedError

    @abstractmethod
    def test_trade(self, coin_in, coin_out, factor):
        raise NotImplementedError

    @abstractmethod
    def make_error_fns(self):
        raise NotImplementedError



"""
Contains order_book function for computing and optionally plotting orderbook
representations of exchange rates between two tokens.
"""

import matplotlib.pyplot as plt
from pandas import DataFrame

from curvesim.pool import CurveMetaPool


def order_book(pool, i, j, *, width=0.1, resolution=10**23, use_fee=True, show=True):
    """
    Computes and optionally plots an orderbook representation of exchange rates
    between two tokens.

    Parameters
    ----------
    pool : CurvePool or CurveMetaPool
        A pool object to compute the bonding curve for.

    i : int
        The index of the "base" token to compute exchange rates for.
        To index the basepool token, use "bp_token"

    j : int
        The index of the "quote" token traded against.
        To index the basepool token, use "bp_token"

    width: float, default=0.1
        The width of the orderbook to compute
        (i.e., from price*(1-width) to price*(1+width))

    resolution : int, default=10**23
        The trade-size interval between points in the orderbook (18-decimal precision)

    use_fee : bool, default=True
        If True, includes fees when computing the orderbook.

    show : bool, default=True
        If true, plots the bonding curve.

    Returns
    -------
    bids : pandas.DataFrame
        DataFrame of prices and depths for each point on the "bid" side of the orderbook

    asks : pandas.DataFrame
        DataFrame of prices and depths for each point on the "ask" side of the orderbook

    """
    snapshot = pool.get_snapshot()
    i, j, functions = _orderbook_args(pool, i, j)
    get_price, exchange = functions

    # Bids
    depth = 0
    price = get_price(i, j, use_fee=use_fee)

    bids = [(price, depth)]

    while bids[-1][0] > bids[0][0] * (1 - width):
        depth += resolution

        dy, _ = exchange(i, j, depth)
        price = get_price(i, j, use_fee=use_fee)

        bids.append((price, depth / 10**18))

        # Return to initial state
        pool.revert_to_snapshot(snapshot)

    # Asks
    depth = 0
    price = get_price(j, i, use_fee=use_fee)

    asks = [(1 / price, depth)]

    while asks[-1][0] < asks[0][0] * (1 + width):
        depth += resolution
        dy, _ = exchange(j, i, depth)
        price = get_price(j, i, use_fee=use_fee)

        asks.append((1 / price, dy / 10**18))

        # Return to initial state
        pool.revert_to_snapshot(snapshot)

    # Format DataFrames
    bids = DataFrame(bids, columns=["price", "depth"]).set_index("price")
    asks = DataFrame(asks, columns=["price", "depth"]).set_index("price")

    if show:
        plt.plot(bids, color="red")
        plt.plot(asks, color="green")
        plt.xlabel("Price")
        plt.ylabel("Depth")
        plt.show()

    return bids, asks


def _orderbook_args(pool, i, j):
    if isinstance(pool, CurveMetaPool):
        # Set functions/parameters
        get_price = pool.dydx
        exchange = pool.exchange_underlying

        # Price function closure if bp_token used
        def get_meta_price(i, j, use_fee):
            # pylint: disable=protected-access
            xp = pool._xp()
            price = pool._dydx(i, j, xp, use_fee=use_fee)
            return price

        if i == "bp_token":
            i = pool.max_coin
            get_price = get_meta_price
            exchange = pool.exchange

        if j == "bp_token":
            j = pool.max_coin
            get_price = get_meta_price
            exchange = pool.exchange

    else:
        get_price = pool.dydx
        exchange = pool.exchange

    return i, j, (get_price, exchange)



"""
Implements the volume-limited arbitrage pipeline.
"""

import os
from functools import partial

from numpy import array, isnan
from scipy.optimize import least_squares, root_scalar

from curvesim.iterators.param_samplers import Grid
from curvesim.iterators.price_samplers import PriceVolume
from curvesim.logging import get_logger
from curvesim.metrics import StateLog, init_metrics, make_results
from curvesim.metrics import metrics as Metrics
from curvesim.pool import get_sim_pool
from curvesim.pool_data.cache import PoolDataCache

from .templates import TradeData, run_pipeline
from .utils import compute_volume_multipliers

logger = get_logger(__name__)


# pylint: disable-next=too-many-arguments
def volume_limited_arbitrage(
    pool_metadata,
    pool_data_cache=None,
    variable_params=None,
    fixed_params=None,
    metrics=None,
    test=False,
    days=60,
    src="coingecko",
    data_dir="data",
    vol_mult=None,
    vol_mode=1,
    ncpu=None,
    end=None,
):
    """
    Implements the volume-limited arbitrage pipeline.

    At each timestep, the pool is arbitraged as close to the prevailing market price
    as possible without surpassing a volume constraint. By default, volume is limited
    to the total market volume at each timestep, scaled by the proportion of
    volume attributable to the pool over the whole simulation period (vol_mult).

    Parameters
    ----------
    pool_metadata : :class:`~curvesim.pool_data.metadata.PoolMetaDataInterface`
        Pool metadata object for the pool of interest.

    variable_params : dict, defaults to broad range of A/fee values
        Pool parameters to vary across simulations.
        keys: pool parameters, values: iterables of ints

        Example
        --------
        >>> variable_params = {"A": [100, 1000], "fee": [10**6, 4*10**6]}

    fixed_params : dict, optional
        Pool parameters set before all simulations.
        keys: pool parameters, values: ints

        Example
        --------
        >>> fixed_params = {"D": 1000000*10**18}

    test : bool, optional
        Overrides variable_params to use four test values:

        .. code-block::

            {"A": [100, 1000], "fee": [3000000, 4000000]}

    days : int, default=60
        Number of days to pull pool and price data for.

    src : str, default="coingecko"
        Source for price/volume data: "coingecko" or "local".

    data_dir : str, default="data"
        relative path to saved price data folder

    vol_mult : float or numpy.ndarray, default computed from data
        Value(s) multiplied by market volume to specify volume limits
        (overrides vol_mode).

        Can be a scalar or vector with values for each pairwise coin combination.

    vol_mode : int, default=1
        Modes for limiting trade volume.

        1: limits trade volumes proportionally to market volume for each pair

        2: limits trade volumes equally across pairs

        3: mode 2 for trades with meta-pool asset, mode 1 for basepool-only trades

    ncpu : int, default=os.cpu_count()
        Number of cores to use.

    Returns
    -------
    dict

    """
    if test:
        variable_params = TEST_PARAMS

    if ncpu is None:
        cpu_count = os.cpu_count()
        ncpu = cpu_count if cpu_count is not None else 1

    variable_params = variable_params or DEFAULT_PARAMS
    metrics = metrics or DEFAULT_METRICS

    if pool_data_cache is None:
        pool_data_cache = PoolDataCache(pool_metadata, days=days, end=end)

    pool = get_sim_pool(pool_metadata, pool_data_cache=pool_data_cache)
    coins = pool_metadata.coins

    param_sampler = Grid(pool, variable_params, fixed_params=fixed_params)
    price_sampler = PriceVolume(coins, days=days, data_dir=data_dir, src=src, end=end)

    if vol_mult is None:
        total_pool_volume = pool_data_cache.volume
        total_market_volume = price_sampler.total_volumes()
        vol_mult = compute_volume_multipliers(
            total_pool_volume,
            total_market_volume,
            pool_metadata.n,
            pool_metadata.pool_type,
            mode=vol_mode,
        )

    metrics = init_metrics(metrics, pool=pool, freq=price_sampler.freq)
    strat = partial(strategy, metrics=metrics, vol_mult=vol_mult)

    output = run_pipeline(param_sampler, price_sampler, strat, ncpu=ncpu)
    results = make_results(*output, metrics)

    return results


# Strategy
def strategy(pool, parameters, price_sampler, metrics, vol_mult):
    """
    Computes and executes volume-limited arbitrage trades at each timestep.

    Parameters
    ----------
    pool : Pool, MetaPool, or RaiPool
        The pool to be arbitraged.

    parameters : dict
        Current pool parameters from the param_sampler (only used for logging/display).

    price_sampler : iterator
        Iterator that returns prices and volumes at each timestep.

    vol_mult : float or numpy.ndarray
        Value(s) multiplied by market volume to specify volume limits.

        Can be a scalar or vector with values for each pairwise coin combination.

    Returns
    -------
    metrics : tuple of lists

    """
    trader = Arbitrageur(pool)
    state_log = StateLog(pool, metrics)

    symbol = pool.symbol
    logger.info(f"[{symbol}] Simulating with {parameters}")

    for price_sample in price_sampler:
        volume_limits = price_sample.volumes * vol_mult
        pool.prepare_for_trades(price_sample.timestamp)
        trade_data = trader.arb_pool(price_sample.prices, volume_limits)
        state_log.update(price_sample=price_sample, trade_data=trade_data)

    return state_log.compute_metrics()


class Arbitrageur:
    """
    Computes, executes, and reports out arbitrage trades.
    """

    def __init__(self, pool):
        """
        Parameters
        ----------
        pool :
            Simulation interface to a subclass of :class:`.Pool`.

        """
        self.pool = pool

    def compute_trades(self, prices, volume_limits):
        """
        Computes trades to optimally arbitrage the pool, constrained by volume limits.

        Parameters
        ----------
        prices : pandas.Series
            Current market prices from the price_sampler.

        volume_limits : pandas.Series
            Current volume limits.

        Returns
        -------
        trades : list of tuples
            List of trades to perform.
            Trades are formatted as (coin_in, coin_out, trade_size).

        errors : numpy.ndarray
            Post-trade price error between pool price and market price.

        res : scipy.optimize.OptimizeResult
            Results object from the numerical optimizer.

        """
        trades, errors, res = opt_arb_multi(self.pool, prices, volume_limits)
        return trades, errors, res

    def do_trades(self, trades):
        """
        Executes a series of trades.

        Parameters
        ----------
        trades : list of tuples
            Trades to execute, formatted as (coin_in, coin_out, trade_size).

        Returns
        -------
        trades_done: list of tuples
            Trades executed, formatted as (coin_in, coin_out, amount_in, amount_out).

        volume : int
            Total volume of trades in 18 decimal precision.

        """
        if len(trades) == 0:
            return [], 0

        total_volume = 0
        trades_done = []
        for trade in trades:
            i, j, dx = trade
            dy, fee, volume = self.pool.trade(i, j, dx)
            trades_done.append((i, j, dx, dy, fee))

            total_volume += volume

        return trades_done, total_volume

    def arb_pool(self, prices, volume_limits):
        trades, price_errors, _ = self.compute_trades(prices, volume_limits)
        trades_done, volume = self.do_trades(trades)
        return TradeData(trades_done, volume, volume_limits, price_errors)


# Optimizers
def opt_arb(get_bounds, error_function, i, j, p):
    """
    Estimates a single trade to optimally arbitrage coin[i] for coin[j] given
    external price p (base: i, quote: j).

    p must be less than dy[j]/dx[i], including fees.

    Parameters
    ----------
    get_bounds : callable
        Function that returns bounds on trade size hypotheses for any
        token pairs (i,j).

    error_function: callable
        Error function that returns the difference between pool price and
        market price (p) after some trade (coin_i, coin_j, trade_size)

    i : int
        Index for the input coin (base)

    j : int
        Index for the output coin (quote)

    p : float
        External market price to arbitrage the pool to

    Returns
    -------
    trade : tuple
        Trades to perform, formatted as (coin_i, coin_j, trade_size).

    error : numpy.ndarray
        Post-trade price error between pool price and market price.

    res : scipy.optimize.OptimizeResult
        Results object from the numerical optimizer.

    """
    bounds = get_bounds(i, j)
    res = root_scalar(error_function, args=(i, j, p), bracket=bounds, method="brentq")

    trade = (i, j, int(res.root))
    error = error_function(int(res.root), i, j, p)

    return trade, error, res


def opt_arb_multi(pool, prices, limits):  # noqa: C901
    """
    Computes trades to optimally arbitrage the pool, constrained by volume limits.

    Parameters
    ----------
    pool :
        Simulation interface to a subclass of :class:`Pool`.

    prices : pandas.Series
        Current market prices from the price_sampler

    volume_limits : pandas.Series
        Current volume limits

    Returns
    -------
    trades : list of tuples
        List of trades to perform.
        Trades are formatted as (coin_i, coin_j, trade_size)

    errors : numpy.ndarray
        Post-trade price error between pool price and market price for each token pair.

    res : scipy.optimize.OptimizeResult
        Results object from the numerical optimizer.

    """
    (
        get_bounds,
        error_function,
        error_function_multi,
        index_combos,
    ) = pool.make_error_fns()
    x0, lo, hi, coins, price_targets = get_trade_args(
        pool.price,
        get_bounds,
        error_function,
        prices,
        limits,
        index_combos,
    )

    # Order trades in terms of expected size
    order = sorted(range(len(x0)), reverse=True, key=x0.__getitem__)
    x0 = [x0[i] for i in order]
    lo = [lo[i] for i in order]
    hi = [hi[i] for i in order]
    coins = [coins[i] for i in order]
    price_targets = [price_targets[i] for i in order]

    # Find trades that minimize difference between
    # pool price and external market price
    trades = []
    try:
        res = least_squares(
            error_function_multi,
            x0=x0,
            args=(price_targets, coins),
            bounds=(lo, hi),
            gtol=10**-15,
            xtol=10**-15,
        )

        # Format trades into tuples, ignore if dx=0
        dxs = res.x

        for k, dx in enumerate(dxs):
            if isnan(dx):
                continue

            dx = int(dx)
            if dx > 0:
                i = coins[k][0]
                j = coins[k][1]
                trades.append((i, j, dx))

        errors = res.fun

    except Exception:
        logger.error(
            f"Optarbs args: x0: {x0}, lo: {lo}, hi: {hi}, prices: {price_targets}",
            exc_info=True,
        )
        errors = array(error_function_multi([0] * len(x0), price_targets, coins))
        res = []

    return trades, errors, res


def get_trade_args(get_pool_price, get_bounds, error_function, prices, limits, combos):
    """
    Returns initial guesses (x0), bounds (lo, hi), ordered coin-pairs, and
    price targets used to estimate the optimal set of arbitrage trades.

    Parameters
    ----------
    get_pool_price : callable
        Function that returns the pool price for any token pair (i,j)

    get_bounds : callable
        Function that returns bounds on trade size hypotheses for
        any token pairs (i,j).

    error_function: callable
        Error function that returns the difference between pool price and
        market price (p) after some trade (coin_i, coin_j, trade_size)

    prices : iterable
        External market prices for each coin-pair

    combos : iterable of tuples
        Ordered pairwise combinations of coin indices

    Returns
    -------
    x0 : list
        Initial "guess" values for trade size for each token pair

    lo : list
        Lower bounds on trade sizes for each token pair

    hi: list
        Upper bounds on trade sizes for each token pair

    coins : list of tuples
        Ordered list of token pairs

    price_targets : list of floats
        Ordered list of price targets for each token pair

    """
    x0 = []
    lo = []
    hi = []
    coins = []
    price_targets = []

    for k, pair in enumerate(combos):
        i, j = pair
        limit = int(limits[k] * 10**18)
        lo.append(0)
        hi.append(limit + 1)

        if get_pool_price(i, j) - prices[k] > 0:
            price = prices[k]
            in_index = i
            out_index = j
        elif get_pool_price(j, i) - 1 / prices[k] > 0:
            price = 1 / prices[k]
            in_index = j
            out_index = i
        else:
            x0.append(0)
            coins.append((i, j))
            price_targets.append(prices[k])
            continue

        try:
            trade, _, _ = opt_arb(
                get_bounds, error_function, in_index, out_index, price
            )
            size = min(trade[2], limit)
            x0.append(size)
        except ValueError:
            pool_price = get_pool_price(in_index, out_index)
            logger.error(
                f"Opt_arb error: Pair: {(in_index, out_index)}, Pool price: {pool_price},"
                f"Target Price: {price}, Diff: {pool_price - price}"
            )
            x0.append(0)

        coins.append((in_index, out_index))
        price_targets.append(price)

    return x0, lo, hi, coins, price_targets


# Defaults
DEFAULT_METRICS = [
    Metrics.Timestamp,
    Metrics.PoolValue,
    Metrics.PoolBalance,
    Metrics.PriceDepth,
    Metrics.ArbMetrics,
]

DEFAULT_PARAMS = {
    "A": [int(2 ** (a / 2)) for a in range(12, 28)],
    "fee": list(range(1000000, 5000000, 1000000)),
}

TEST_PARAMS = {"A": [100, 1000], "fee": [3000000, 4000000]}



from abc import abstractmethod, abstractproperty


class PoolMetaDataInterface:
    @abstractmethod
    def init_kwargs(self, balanced=True, balanced_base=True, normalize=True):
        raise NotImplementedError

    @abstractproperty
    def address(self):
        raise NotImplementedError

    @abstractproperty
    def chain(self):
        raise NotImplementedError

    @abstractproperty
    def pool_type(self):
        """
        Returns the pool type.

        Returns
        -------
        Pool

        """
        raise NotImplementedError

    @abstractproperty
    def sim_pool_type(self):
        """
        Returns the pool type.

        Returns
        -------
        SimPool

        """
        raise NotImplementedError

    @abstractproperty
    def coins(self):
        """
        Returns coin addresses for the pool's holdings.

        For pools that are not on Ethereum mainnet, the address
        for the corresponding mainnet token is returned.

        For lending tokens (e.g., aTokens or cTokens), the
        address for the underlying token is returned.

        Returns
        -------
        list of strings
            coin addresses

        """
        raise NotImplementedError

    @abstractproperty
    def coin_names(self):
        """
        Returns coin names for the pool's holdings.

        For pools that are not on Ethereum mainnet, the name
        of the corresponding mainnet token is returned.

        For lending tokens (e.g., aTokens or cTokens), the
        name of the underlying token is returned.

        Returns
        -------
        list of strings
            coin names

        """
        raise NotImplementedError

    @abstractproperty
    def n(self):
        """
        Returns the number of token-types (e.g., DAI, USDC, USDT) in a pool.

        Returns
        -------
        int or list of ints
            Number of token-types.

            For metapools, a list [n_metapool, n_basepool] is returned.

            N_metapool includes the basepool LP token.

        """
        raise NotImplementedError


class PoolMetaDataBase(PoolMetaDataInterface):
    def __init__(self, metadata_dict, pool_type, sim_pool_type):
        self._dict = metadata_dict
        self._pool_type = pool_type
        self._sim_pool_type = sim_pool_type

    @property
    def address(self):
        return self._dict["address"]

    @property
    def chain(self):
        return self._dict["chain"]

    @property
    def pool_type(self):
        return self._pool_type

    @property
    def sim_pool_type(self):
        return self._sim_pool_type



from altair import TitleParams, vconcat

from curvesim.plot.result_plotter import ResultPlotter
from curvesim.utils import override

from .make_page import make_page_from_results


def plot_results(results):
    """
    Takes a :class:`.metrics.SimResults` object and plots the data returned by both
    :func:`SimResults.summary` and :func:`SimResults.data`
    """

    summary = plot_summary(results)
    timeseries = plot_data(results)
    return vconcat(summary, timeseries).resolve_scale(color="independent")


def plot_summary(results):
    """
    Takes a :class:`.metrics.SimResults` object and plots the data returned by
    :func:`SimResults.summary`.
    """
    title = TitleParams(text="Summary Metrics", fontSize=16)
    data_key = "summary"
    axes = {"metric": "y", "dynamic": {"x": "x:Q", "color": "color:O"}}

    page = make_page_from_results(results, data_key, axes)
    return page.properties(title=title)


def plot_data(results):
    """
    Takes a :class:`.metrics.SimResults` object and plots the data returned by
    :func:`SimResults.data`.
    """
    title = TitleParams(text="Timeseries Data", fontSize=16)
    data_key = "data"
    axes = {"metric": "y", "dynamic": {"color": "color:O"}}

    page = make_page_from_results(results, data_key, axes, downsample=True)
    return page.properties(title=title)


class AltairResultPlotter(ResultPlotter):
    """
    :class:`.plot.ResultPlotter` implementation using Altair.
    """

    @override
    def save(self, chart, save_as):
        chart.save(save_as)


result_plotter = AltairResultPlotter(plot_data, plot_results, plot_summary)



from altair import Axis, Scale

STYLES = {
    "line": {
        "mark": {"type": "line"},
        "encoding": {
            "x": {"scale": Scale(zero=False)},
            "y": {"scale": Scale(zero=False)},
        },
    },
    "point_line": {
        "mark": {"type": "line", "point": True},
        "encoding": {
            "x": {"scale": Scale(zero=False)},
            "y": {"scale": Scale(zero=False)},
        },
    },
    "time_series": {
        "mark": {"type": "line"},
        "encoding": {
            "x": {"title": None, "shorthand": "timestamp"},
            "y": {"scale": Scale(zero=False)},
        },
    },
    "histogram": {
        "mark": {"type": "line", "interpolate": "step-before"},
        "encoding": {
            "x": {"scale": Scale(zero=False)},
            "y": {
                "title": "Frequency",
                "shorthand": "frequency",
                "scale": Scale(zero=True),
                "axis": Axis(format="%"),
            },
        },
    },
}



"""
Pool subpackage for creating Curve pool objects.

This subpackage can be used standalone from the simulation functionality
for interactive exploration of Curve pool behavior. The two primary ways to create
pools are `get` and `make`, which allow you to create pools from on-chain state or p
assed-in parameters, respectively.

Pool calculations are thoroughly unit-tested against the corresponding
smart contract code.  We aim for the exact results as in the vyper
integer arithmetic. The pool interfaces largely adhere to the smart contracts
but in a few cases allow an extra option, such as enabling/disabling fees.
"""

__all__ = [
    "get",
    "make",
    "Pool",
    "CurvePool",
    "CurveMetaPool",
    "CurveRaiPool",
    "CurveCryptoPool",
    "SimCurvePool",
    "SimCurveMetaPool",
    "SimCurveRaiPool",
]

from curvesim.exceptions import CurvesimValueError
from curvesim.pool_data import get_metadata
from curvesim.pool_data.metadata import PoolMetaData, PoolMetaDataInterface

from .base import Pool
from .cryptoswap import CurveCryptoPool
from .sim_interface import SimCurveMetaPool, SimCurvePool, SimCurveRaiPool
from .stableswap import CurveMetaPool, CurvePool, CurveRaiPool


def make(
    A,
    D,
    n,
    basepool=None,
    rates=None,
    rate_multiplier=None,
    tokens=None,
    fee=4 * 10**6,
    fee_mul=None,
    admin_fee=0 * 10**9,
):
    """
    Factory function for creating pools from "raw" parameters, i.e. no
    data pulled from chain.

    Parameters
    ----------
    A: int
        Amplification coefficient.  This controls the curvature of the
        stableswap bonding curve.  Increased values makes the curve
        flatter in a greater neighborhood of equal balances.

        Defaults to [int(2 ** (a / 2)) for a in range(12, 28)].

    D: int
        Total pool liquidity given in 18 decimal precision.

    n: int
        The number of token-types in the pool (e.g., DAI, USDC, USDT = 3)

    basepool: dict, optional
        a dict cointaining the arguments for instantiating a basepool

    rates: list of int, optional
        precisions for each coin

    rate_multiplier: int, optional
        precision and rate adjustment for primary stable in metapool

    tokens: int, optional
        Total LP token supply.

        .. note::
            The number of tokens does not influence "normal" pool computations;
            but, for metapools, the number of basepool tokens is critically
            important to trade calculations.

    fee: int
        Fees taken for both liquidity providers and the DAO.

        Units are in fixed-point so that 10**10 is 100%,
        e.g. 4 * 10**6 is 4 bps and 2 * 10**8 is 2%.

    fee_mul : int
        fee multiplier for dynamic fee pools

    admin_fee : int, default=0 * 10**9
        Fees taken for the DAO.  For factory pools, it is half of the total fees,
        as was typical for previous non-factory pools.

        Units are fixed-point percentage of `fee`, e.g. 5 * 10**9 is
        50% of the total fees.

    Returns
    -------
    :class:`Pool`
    """
    if rates and rate_multiplier:
        raise CurvesimValueError("Should have only `rates` or `rate_multiplier`.")

    if basepool:
        pool = CurveMetaPool(
            A,
            D,
            n,
            basepool,
            rate_multiplier=rate_multiplier,
            tokens=tokens,
            fee=fee,
            fee_mul=fee_mul,
            admin_fee=admin_fee,
        )

    else:
        pool = CurvePool(
            A,
            D,
            n,
            rates=rates,
            tokens=tokens,
            fee=fee,
            fee_mul=fee_mul,
            admin_fee=admin_fee,
        )

    return pool


def get_pool(
    pool_metadata,
    chain="mainnet",
    balanced=False,
    balanced_base=False,
    normalize=False,
):
    """
    Constructs a pool object based on the stored data.

    Parameters
    ----------
    address : str
        pool address prefixed with "0x"

    chain: str, default="mainnet"
        chain/layer2 identifier, e.g. "mainnet", "arbitrum", "optimism"

    balanced : bool, default=False
        If True, balances the pool value across assets.

    balanced_base : bool, default=False
        If True and pool is metapool, balances the basepool value across assets.

    normalize : bool, default=False
        If True, normalizes balances to 18 decimals (useful for sim calculations).

    sim: bool, default=False
        If True, returns a `SimPool` version of the pool.

    Returns
    -------
    :class:`Pool`

    Examples
    --------
    >>> import curvesim
    >>> pool = curvesim.pool.get("0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7", "mainnet")

    """
    if isinstance(pool_metadata, str):
        pool_metadata = get_metadata(pool_metadata, chain=chain)
    elif isinstance(pool_metadata, dict):
        pool_metadata = PoolMetaData(pool_metadata)
    elif isinstance(pool_metadata, PoolMetaDataInterface):
        pass
    else:
        raise CurvesimValueError(
            "`pool_metadata` must be of type `str`, `dict`, or `PoolMetaDataInterface`."
        )

    init_kwargs = pool_metadata.init_kwargs(balanced, balanced_base, normalize)

    pool_type = pool_metadata.pool_type
    pool = pool_type(**init_kwargs)

    pool.metadata = pool_metadata._dict  # pylint: disable=protected-access

    return pool


POOL_TYPE_TO_CUSTOM_KWARGS = {SimCurveRaiPool: ["redemption_prices"]}


def get_sim_pool(
    pool_metadata,
    chain="mainnet",
    balanced=True,
    balanced_base=True,
    custom_kwargs=None,
    pool_data_cache=None,
):
    """
    Effectively the same as the `get_pool` function but returns
    an object in the `SimPool` hierarchy.
    """
    custom_kwargs = custom_kwargs or {}

    if isinstance(pool_metadata, str):
        pool_metadata = get_metadata(pool_metadata, chain=chain)
    elif isinstance(pool_metadata, dict):
        pool_metadata = PoolMetaData(pool_metadata)
    elif isinstance(pool_metadata, PoolMetaDataInterface):
        pass
    else:
        raise CurvesimValueError(
            "`pool_metadata` must be of type `str`, `dict`, or `PoolMetaDataInterface`."
        )

    init_kwargs = pool_metadata.init_kwargs(balanced, balanced_base, normalize=True)

    pool_type = pool_metadata.sim_pool_type

    custom_keys = POOL_TYPE_TO_CUSTOM_KWARGS.get(pool_type, [])
    for key in custom_keys:
        try:
            init_kwargs[key] = custom_kwargs.get(key) or getattr(pool_data_cache, key)
        except KeyError as e:
            raise CurvesimValueError(f"'{pool_type.__name__}' needs '{key}'.") from e

    pool = pool_type(**init_kwargs)

    pool.metadata = pool_metadata._dict  # pylint: disable=protected-access

    return pool


get = get_pool



from altair import (
    Color,
    CalculateTransform,
    FilterTransform,
    Scale,
    concat,
    hconcat,
)

from ..selectors import make_selector


def make_result_selectors(factors, dynamic_axes):
    dynamic_axes = dict(list(dynamic_axes.items())[: len(factors)])

    axes = make_axis_selectors(factors, dynamic_axes)
    parameters = make_parameter_filters(factors, dynamic_axes)

    # kwargs to pass to metric charts
    metric_kwargs = {
        axis: {"shorthand": shorthand, "title": None}
        for axis, shorthand in dynamic_axes.items()
    }
    metric_kwargs["transform"] = axes["transforms"] + parameters["transforms"]

    return {
        "kwargs": metric_kwargs,
        "charts": format_selector_charts(axes["charts"], parameters["charts"]),
    }


def make_axis_selectors(factors, dynamic_axes):
    factor_names = list(factors.keys())

    charts = []
    transforms = []
    for idx, axis in enumerate(dynamic_axes):
        chart, transform = _make_axis_selector(axis, factor_names, idx)

        charts.append(chart)
        transforms.append(transform)

    return {"charts": charts, "transforms": transforms}


def make_parameter_filters(factors, dynamic_axes):
    n_dynamic_axes = len(dynamic_axes)

    charts = []
    transforms = []
    for i, factor in enumerate(factors.items()):
        idx = "all" if i < n_dynamic_axes else 0
        chart, transform = _make_parameter_filter(*factor, idx)

        charts.append(chart)
        transforms.append(transform)

    return {"charts": charts, "transforms": transforms}


def _make_axis_selector(axis, options, sel_idx):
    chart, selector = make_selector(
        axis,
        options,
        title=axis,
        style={"axis": "y", "orient": "left"},
        toggle=False,
        sel_idx=sel_idx,
    )

    calc_kwargs = {"calculate": f"datum[{selector.name}.{axis}]", "as": axis}
    return chart, CalculateTransform(**calc_kwargs)


def _make_parameter_filter(factor, options, sel_idx):
    color = Color("labels:O", scale=Scale(scheme="viridis"), legend=None)

    chart, selector = make_selector(
        factor,
        options,
        title=factor,
        style={"axis": "x", "orient": "bottom", "color": color},
        toggle="true",
        sel_idx=sel_idx,
    )
    return chart, FilterTransform(selector)


def format_selector_charts(axis_selector_charts, parameter_filter_charts):
    left = concat(*axis_selector_charts, title="Axis Selectors:")
    right = concat(*parameter_filter_charts, title="Toggle Filters:")
    return hconcat(left, right.resolve_scale(color="independent"))



class CurvesimException(Exception):
    """Base exception class"""


class SubgraphError(CurvesimException):
    """Raised for errors in subgraph query"""


class SubgraphResultError(SubgraphError):
    """Raised when subgraph results aren't as expected"""


class MissingEnvVarError(CurvesimException, RuntimeError):
    """Environment variable is missing."""


class HttpClientError(CurvesimException):
    """Raised for errors from async HTTP client request."""

    def __init__(self, status, message, url=None):
        super().__init__(status, message)
        self.status = status
        self.message = message
        self.url = url

    def __repr__(self):
        return f"HttpClientError({self.status}, {self.message}, url={self.url})"


class CurvesimValueError(CurvesimException, ValueError):
    """Raised when an argument has an inappropriate value (but the right type)."""


class SnapshotError(CurvesimException):
    """Error using a snapshot."""


class CalculationError(CurvesimException):
    """Error during a mathematical calculation."""


class CryptoPoolError(RuntimeError):
    """Runtime error from a CryptoSwap pool."""


class MetricError(CurvesimException):
    """Error using a metric."""


class ResultsError(CurvesimException):
    """Error processing pipeline/simulation results."""


class PlotError(CurvesimException):
    """Error in plotting."""


class NetworkError(CurvesimException):
    """Error for network subpackage."""



from curvesim.pool.stableswap.pool import CurvePool

from .base import PoolMetaDataBase


class StableswapMetaData(PoolMetaDataBase):
    def init_kwargs(self, balanced=True, balanced_base=True, normalize=True):
        data = self._dict

        def process_to_kwargs(data, balanced, normalize):
            kwargs = {
                "A": data["params"]["A"],
                "D": data["reserves"]["D"],
                "n": len(data["coins"]["names"]),
                "fee": data["params"]["fee"],
                "fee_mul": data["params"]["fee_mul"],
                "tokens": data["reserves"]["tokens"],
            }
            if not normalize:
                if data["basepool"]:
                    d = data["coins"]["decimals"][0]
                    kwargs["rate_multiplier"] = 10 ** (36 - d)
                else:
                    kwargs["rates"] = [
                        10 ** (36 - d) for d in data["coins"]["decimals"]
                    ]
            if not balanced:
                if normalize:
                    coin_balances = data["reserves"]["by_coin"]
                else:
                    coin_balances = data["reserves"]["unnormalized_by_coin"]
                kwargs["D"] = coin_balances
            return kwargs

        kwargs = process_to_kwargs(data, balanced, normalize)

        if data["basepool"]:
            bp_data = data["basepool"]
            bp_kwargs = process_to_kwargs(bp_data, balanced_base, normalize)
            basepool = CurvePool(**bp_kwargs)
            basepool.metadata = bp_data
            kwargs["basepool"] = basepool

        return kwargs

    @property
    def coins(self):
        if not self._dict["basepool"]:
            c = self._dict["coins"]["addresses"]
        else:
            c = (
                self._dict["coins"]["addresses"][:-1]
                + self._dict["basepool"]["coins"]["addresses"]
            )
        return c

    @property
    def coin_names(self):
        if not self._dict["basepool"]:
            c = self._dict["coins"]["names"]
        else:
            c = (
                self._dict["coins"]["names"][:-1]
                + self._dict["basepool"]["coins"]["names"]
            )
        return c

    @property
    def n(self):
        data = self._dict
        if data["basepool"]:
            n = [
                len(data["coins"]["names"]),
                len(data["basepool"]["coins"]["names"]),
            ]
        else:
            n = len(data["coins"]["names"])

        return n



from datetime import datetime, timedelta, timezone

from curvesim.logging import get_logger
from curvesim.network import coingecko as _coingecko
from curvesim.network import nomics as _nomics

logger = get_logger(__name__)


def nomics(coins, days=60, data_dir="data", end=None):
    if end is None:
        t_end = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)
        custom_suffix = ""
    else:
        t_end = datetime.fromtimestamp(end, tz=timezone.utc)
        custom_suffix = "-" + str(end)
    t_start = t_end - timedelta(days=days)

    logger.info("Fetching Nomics price data...")
    logger.info("Timerange: %s to %s" % (str(t_start), str(t_end)))

    _nomics.update(
        coins, None, t_start, t_end, data_dir=data_dir, custom_suffix=custom_suffix
    )
    prices, volumes, pzero = _nomics.pool_prices(
        coins, data_dir=data_dir, custom_suffix=custom_suffix
    )

    return prices, volumes, pzero


def coingecko(coins, days=60):
    logger.info("Fetching CoinGecko price data...")
    prices, volumes = _coingecko.pool_prices(coins, "usd", days)
    pzero = 0

    return prices, volumes, pzero


def local(coins, data_dir="data", end=None):
    logger.info("Using local data...")
    if end is not None:
        t_end = datetime.fromtimestamp(end, tz=timezone.utc)
        custom_suffix = "-" + str(end)
    else:
        t_end = None
        custom_suffix = ""
    prices, volumes, pzero = _nomics.local_pool_prices(
        coins, data_dir=data_dir, t_end=t_end, custom_suffix=custom_suffix
    )

    return prices, volumes, pzero



from curvesim.pool.sim_interface import SimCurveMetaPool, SimCurvePool, SimCurveRaiPool


def get_pool_parameters(pool):
    """
    Returns pool parameters for the input pool. Functions for each pool type are
    specified in the `pool_parameter_functions` dict.
    """
    return pool_parameter_functions[type(pool)](pool)


def get_stableswap_pool_params(pool):
    """Returns pool parameters for stableswap non-meta pools."""
    params = {
        "A": pool.A,
        "D": pool.D() / 10**18,
        "fee": pool.fee / 10**10,
    }

    for param_name in ["fee_mul", "admin_fee"]:
        param_val = getattr(pool, param_name, None)
        if param_val:
            params.update({param_name: param_val / 10**10})

    return params


def get_stableswap_metapool_params(pool):
    """Returns pool parameters for stableswap metapools."""
    pool_params = get_stableswap_pool_params(pool)
    bp_params = get_stableswap_pool_params(pool.basepool)
    bp_params = {key + "_base": val for key, val in bp_params.items()}
    pool_params.update(bp_params)
    return pool_params


pool_parameter_functions = {
    SimCurvePool: get_stableswap_pool_params,
    SimCurveMetaPool: get_stableswap_metapool_params,
    SimCurveRaiPool: get_stableswap_metapool_params,
}



"""
Contains bonding_curve function for computing and optionally plotting a pool's
bonding curve and current reserves.
"""
from itertools import combinations

import matplotlib.pyplot as plt
from numpy import linspace

from curvesim.pool import CurveMetaPool


def bonding_curve(pool, *, truncate=0.0005, resolution=1000, show=True):
    """
    Computes and optionally plots a pool's bonding curve and current reserves.

    Parameters
    ----------
    pool : CurvePool or CurveMetaPool
        A pool object to compute the bonding curve for.

    truncate : float, default=0.0001
        Determines where to truncate the bonding curve (i.e., D*truncate).

    resolution : int, default=1000
        Number of points along the bonding curve to compute.

    show : bool, default=True
        If true, plots the bonding curve.

    Returns
    -------
    xs : list of lists
        Lists of reserves for the first coin in each combination of token pairs

    ys : list of lists
        Lists of reserves for the second coin in each combination of token pairs

    """

    if isinstance(pool, CurveMetaPool):
        combos = [(0, 1)]
    else:
        combos = list(combinations(range(pool.n), 2))

    try:
        labels = pool.metadata["coins"]["names"]
    except (AttributeError, KeyError):
        labels = [f"Coin {str(label)}" for label in range(pool.n)]

    if show:
        _, axs = plt.subplots(1, len(combos), constrained_layout=True)

    D = pool.D()
    xp = pool._xp()  # pylint: disable=protected-access

    xs_out = []
    ys_out = []
    for n, combo in enumerate(combos):
        i, j = combo

        truncated_D = int(D * truncate)
        xs_n = linspace(
            truncated_D, pool.get_y(j, i, truncated_D, xp), resolution
        ).round()

        ys_n = []
        for x in xs_n:
            ys_n.append(pool.get_y(i, j, int(x), xp))

        xs_n = [x / 10**18 for x in xs_n]
        ys_n = [y / 10**18 for y in ys_n]
        xs_out.append(xs_n)
        ys_out.append(ys_n)

        if show:
            if len(combos) == 1:
                ax = axs
            else:
                ax = axs[n]

            ax.plot(xs_n, ys_n, color="black")
            ax.scatter(xp[i] / 10**18, xp[j] / 10**18, s=40, color="black")
            ax.set_xlabel(labels[i])
            ax.set_ylabel(labels[j])

    if show:
        plt.show()

    return xs_out, ys_out



"""
Metrics recorded during simulations. The submodule includes:

    1. **Metric objects:**
        Objects that compute and record metrics/statistics throughout simulations,
        and provide a config attribute that specifies how to compute, summarize, and/or
        plot the recorded data.

        :mod:`.metrics.base` contains the metrics base class and generic metric classes.

        :mod:`.metrics.metrics` contains specific metric classes for use in simulations.

    2. :class:`.StateLog`:
        Logger that records simulation/pool state throughout a simulation and computes
        metrics at the end of each simulation run.

    3. :class:`.SimResults`:
        Container for simulation results with methods to plot or return recorded
        metrics as DataFrames.
"""

__all__ = ["SimResults", "StateLog", "init_metrics", "make_results", "metrics"]

from .state_log import StateLog
from .results import SimResults, make_results
from . import metrics


def init_metrics(metric_classes, **kwargs):
    """
    Initializes metric classes with **kwargs as keyword arguments. Each metric class
    only uses the keyword arguments specified in its __init__ method.

    Parameters
    ----------
    metric_classes : list
        A list of metric classes (e.g., from curvesim.metrics.metrics)

    kwargs :
        The keyword arguments used when initializing each metric class.

    Returns
    -------
    metrics :
        A list of the initialized metrics.
    """
    return [Metric(**kwargs) for Metric in metric_classes]



"""
Tools for implementing and running simulation pipelines.

Pipelines iterate over pools with parameters set by
:mod:`.param_samplers` and time-series data produced by :mod:`.price_samplers`.
Each pipeline implements a "strategy" dictating what is done at each timestep
(e.g., see :func:`.arbitrage.strategy`).

Typically, a pipeline takes in :class:`.PoolData`; specifies a param_sampler,
price_sampler, and strategy; and returns metrics/results.
"""

__all__ = ["arbitrage", "templates", "utils"]



from pandas import concat

from .sim_results import SimResults


def make_results(data_per_run, data_per_trade, summary_data, metrics):
    """Initializes a results object from the output of a simulation pipeline."""

    for run_n, data in enumerate(data_per_trade):  # add "run" column
        data.insert(0, "run", run_n)

    data_per_run = concat(data_per_run, ignore_index=True)
    data_per_trade = concat(data_per_trade, ignore_index=True)
    summary_data = concat(summary_data, ignore_index=True)

    factors = get_factors(data_per_run)
    plot_config = combine_plot_configs(metrics)

    return SimResults(data_per_run, data_per_trade, summary_data, factors, plot_config)


def combine_plot_configs(metrics):
    """Combines plot configs across all metrics."""

    metrics_config = {}
    summary_config = {}
    for metric in metrics:
        if metric.plot_config:
            metrics = metric.plot_config.get("metrics", {})
            summary = metric.plot_config.get("summary", {})
            metrics_config.update(metrics)
            summary_config.update(summary)
    return {"data": metrics_config, "summary": summary_config}


def get_factors(data_per_run):
    """Returns dict mapping variable parameters to their values."""

    factors = {}
    for parameter in data_per_run:
        unique_values = data_per_run[parameter].unique()
        if len(unique_values) > 1:
            factors[parameter] = list(unique_values)

    return factors



"""Simple health check for the package."""
import time

from .sim import autosim


def hello_world():
    """Simple sim run as a health check."""
    # pylint: disable=redefined-outer-name
    t = time.time()
    res = autosim("0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7", test=True, ncpu=1)
    elapsed = time.time() - t
    print("Elapsed time:", elapsed)

    return res


if __name__ == "__main__":
    res = hello_world()



"""
Submodule for Curve stableswap pools, including Pool, Metapool, and RaiPool classes.

"""

__all__ = [
    "CurveMetaPool",
    "CurvePool",
    "CurveRaiPool",
]


from .metapool import CurveMetaPool
from .pool import CurvePool
from .raipool import CurveRaiPool



from altair import Tooltip


def make_tooltip(encoding, metric_axis, factors, prefix=None):
    tooltip = []
    if "timestamp" in encoding["x"]["shorthand"]:
        tooltip.append(Tooltip(encoding["x"]["shorthand"], title="Time"))

    title = encoding[metric_axis]["title"]
    if prefix:
        title = f"{prefix} {title}"

    tooltip.append(Tooltip(encoding[metric_axis]["shorthand"], title=title))

    return tooltip + factors



from itertools import combinations

from numpy import isnan

from curvesim.exceptions import CurvesimValueError
from curvesim.pool.sim_interface.simpool import SimStableswapBase
from curvesim.pool.stableswap.metapool import CurveMetaPool
from curvesim.utils import override


class SimCurveMetaPool(SimStableswapBase, CurveMetaPool):
    @property
    def _precisions(self):
        p_base = self.basepool.rates[:]
        return [self.rate_multiplier, *p_base]

    @override
    def _init_coin_indices(self):
        meta_coin_names = self.coin_names[:-1]
        base_coin_names = self.basepool.coin_names
        bp_token_name = self.coin_names[-1]

        # indexing from primary stable, through basepool underlyers,
        # and then basepool LP token.
        coin_names = [*meta_coin_names, *base_coin_names, bp_token_name]
        coin_dict = {name: i for i, name in enumerate(coin_names)}

        return coin_dict

    @override
    def price(self, coin_in, coin_out, use_fee=True):
        i, j = self.get_coin_indices(coin_in, coin_out)
        bp_token_index = self.n_total

        if bp_token_index not in (i, j):
            return self.dydx(i, j, use_fee=use_fee)

        if i == bp_token_index:
            i = self.max_coin

        if j == bp_token_index:
            j = self.max_coin

        if i == j:
            raise CurvesimValueError("Duplicate coin indices.")

        xp = self._xp()
        return self._dydx(i, j, xp=xp, use_fee=use_fee)

    @override
    def trade(self, coin_in, coin_out, size):
        """
        Trade between two coins in a pool.
        Coin index runs over basepool underlyers.
        We count only volume when one coin is the primary stable.
        """
        i, j = self.get_coin_indices(coin_in, coin_out)

        out_amount, fee = self.exchange_underlying(i, j, size)

        max_coin = self.max_coin
        if i < max_coin or j < max_coin:
            volume = size * self._precisions[i] // 10**18  # in D units
        else:
            volume = 0

        return out_amount, fee, volume

    @override
    def test_trade(self, coin_in, coin_out, factor, use_fee=True):
        """
        Trade between top-level coins but leaves balances affected.

        Used to compute liquidity density.
        """
        i, j = self.get_coin_indices(coin_in, coin_out)
        bp_token_index = self.n_total
        if bp_token_index not in (i, j):
            raise CurvesimValueError("Must be trade with basepool token.")

        max_coin = self.max_coin

        if i == bp_token_index:
            i = max_coin

        if j == bp_token_index:
            j = max_coin

        if i == j:
            raise CurvesimValueError("Duplicate coin indices.")

        dx = self.balances[i] // factor

        with self.use_snapshot_context():
            self.exchange(i, j, dx)

            xp_post = self._xp()
            price = self._dydx(i, j, xp_post, use_fee=use_fee)

        return price

    @override
    def make_error_fns(self):  # noqa: C901
        # Note: for performance, does not support string coin-names

        max_coin = self.max_coin

        p_all = [self.rate_multiplier] + self.basepool.rates
        xp_meta = self._xp_mem(self.balances, self.rates)
        xp_base = self._xp_mem(self.basepool.balances, self.basepool.rates)

        all_idx = range(self.n_total)
        index_combos = combinations(all_idx, 2)

        def get_trade_bounds(i, j):
            base_i = i - max_coin
            base_j = j - max_coin
            meta_i = max_coin
            meta_j = max_coin
            if base_i < 0:
                meta_i = i
            if base_j < 0:
                meta_j = j

            if base_i < 0 or base_j < 0:
                xp_j = int(xp_meta[meta_j] * 0.01)
                high = self.get_y(meta_j, meta_i, xp_j, xp_meta)
                high -= xp_meta[meta_i]
            else:
                xp_j = int(xp_base[base_j] * 0.01)
                high = self.basepool.get_y(base_j, base_i, xp_j, xp_base)
                high -= xp_base[base_i]

            return (0, high)

        def post_trade_price_error(dx, i, j, price_target):
            dx = int(dx) * 10**18 // p_all[i]

            with self.use_snapshot_context():
                if dx > 0:
                    self.exchange_underlying(i, j, dx)

                dydx = self.dydxfee(i, j)

            return dydx - price_target

        def post_trade_price_error_multi(dxs, price_targets, coins):
            with self.use_snapshot_context():
                # Do trades
                for k, pair in enumerate(coins):
                    i, j = pair

                    if isnan(dxs[k]):
                        dx = 0
                    else:
                        dx = int(dxs[k]) * 10**18 // p_all[i]

                    if dx > 0:
                        self.exchange_underlying(i, j, dx)

                # Record price errors
                errors = []
                for k, pair in enumerate(coins):
                    i, j = pair
                    dydx = self.dydxfee(i, j)
                    errors.append(dydx - price_targets[k])

            return errors

        return (
            get_trade_bounds,
            post_trade_price_error,
            post_trade_price_error_multi,
            index_combos,
        )



from curvesim.pool.sim_interface import SimCurveMetaPool, SimCurvePool, SimCurveRaiPool


def get_pool_state(pool):
    """
    Returns pool state for the input pool. Functions for each pool type are
    specified in the `pool_state_functions` dict.
    """
    return pool_state_functions[type(pool)](pool)


def get_stableswap_pool_state(pool):
    """Returns pool state for stableswap non-meta pools."""
    return {
        "balances": pool.balances,
        "tokens": pool.tokens,
        "admin_balances": pool.admin_balances,
    }


def get_stableswap_metapool_state(pool):
    """Returns pool state for stableswap non-meta pools."""
    state = get_stableswap_pool_state(pool)
    bp_state = get_stableswap_pool_state(pool.basepool)
    bp_state = {key + "_base": val for key, val in bp_state.items()}
    state.update(bp_state)
    return state


pool_state_functions = {
    SimCurvePool: get_stableswap_pool_state,
    SimCurveMetaPool: get_stableswap_metapool_state,
    SimCurveRaiPool: get_stableswap_metapool_state,
}



"""
Canonical source of truth for version info.

The version number should not be changed manually here.  Instead,
a tool such as `bump2version` should be used.
"""
__version__ = "0.4.0.a2"
# FIXME: this logic isn't quite correct here for pre-release versions
__version_info__ = tuple(
    int(num) if num.isdigit() else num for num in __version__.split(".")
)



import asyncio

from ..network.coingecko import crosschain_coin_addresses_sync
from ..network.subgraph import pool_snapshot_sync, symbol_address_sync
from ..network.web3 import underlying_coin_addresses_sync


def from_address(address, chain):
    """
    Returns

    Parameters
    ----------
    address: str
        Address prefixed with '0x'
    chain: str
        Chain name

    Returns
    -------
    Pool snapshot dictionary in the format returned by
    :func:`curvesim.network.subgraph.pool_snapshot`.
    """
    loop = asyncio.get_event_loop()
    data = pool_snapshot_sync(address, chain, event_loop=loop)

    # Get mainnet addresses for coins
    m_addrs = crosschain_coin_addresses_sync(
        data["coins"]["addresses"], chain, "mainnet", event_loop=loop
    )

    data["coins"]["addresses"] = m_addrs

    # Get underlying token addresses
    if data["pool_type"] == "LENDING":
        u_addrs = underlying_coin_addresses_sync(m_addrs, event_loop=loop)

        m = data.pop("coins")
        names = [n[1:] for n in m["names"]]

        data["coins"] = {"names": names, "addresses": u_addrs, "wrapper": m}

    return data


def from_symbol(symbol, chain):
    address = symbol_address_sync(symbol, chain)

    data = from_address(address, chain)

    return data



from .base import PoolMetaDataBase


class CryptoswapMetaData(PoolMetaDataBase):
    def init_kwargs(self, balanced=True, balanced_base=True, normalize=True):
        data = self._dict

        def process_to_kwargs(data, balanced, normalize):
            kwargs = {
                "A": data["params"]["A"],
                "gamma": data["params"]["gamma"],
                "n": len(data["coins"]["names"]),
                "D": data["reserves"]["D"],
                "mid_fee": data["params"]["mid_fee"],
                "out_fee": data["params"]["out_fee"],
                "allowed_extra_profit": data["params"]["allowed_extra_profit"],
                "fee_gamma": data["params"]["fee_gamma"],
                "adjustment_step": data["params"]["adjustment_step"],
                "price_scale": data["params"]["price_scale"],
                # FIXME: when subgraph is fixed, we can pull this
                # "admin_fee": data["params"]["admin_fee"],
                "ma_half_time": data["params"]["ma_half_time"],
                "tokens": data["reserves"]["tokens"],
                "xcp_profit": data["params"]["xcp_profit"],
                "xcp_profit_a": data["params"]["xcp_profit_a"],
            }
            if not normalize:
                kwargs["precisions"] = [
                    10 ** (18 - d) for d in data["coins"]["decimals"]
                ]
            if not balanced:
                if normalize:
                    coin_balances = data["reserves"]["by_coin"]
                else:
                    coin_balances = data["reserves"]["unnormalized_by_coin"]
                kwargs["balances"] = coin_balances
            return kwargs

        kwargs = process_to_kwargs(data, balanced, normalize)

        return kwargs

    @property
    def coins(self):
        return self._dict["coins"]["addresses"]

    @property
    def coin_names(self):
        return self._dict["coins"]["names"]

    @property
    def n(self):
        return len(self._dict["coins"]["names"])



from pandas import Grouper, Series, concat

from curvesim.exceptions import PlotError


def preprocess_data(data, config, factors, downsample=False):
    if downsample:
        return downsample_data(data, config, factors)
    else:
        return {"main": data.round(6)}


def downsample_data(data, config, factors):
    data_out = {}
    resample = {"run": "last"}
    for metric, cfg in config.items():
        rule = cfg.get("resample")

        if rule:
            resample[metric] = rule

        elif rule is False:
            data_out[metric] = data.pop(metric)

        elif cfg["style"].startswith("histogram"):
            d = concat([data[factors], data.pop(metric)], axis=1)
            data_out[metric] = to_histograms(d, factors)

        else:
            raise PlotError(f"No resample strategy found in {metric} config.")

    data_out["main"] = resample_data(data, factors, resample)
    return data_out


def resample_data(data, factors, fns):
    groups = factors + [Grouper(freq="1D", key="timestamp")]
    resampled = data.groupby(groups).agg(fns).round(6)
    return resampled.reset_index()


def to_histograms(data, factors):
    return data.groupby(factors).apply(make_histogram).reset_index()


def make_histogram(data, bins=500):
    metric = data.iloc[:, -1]
    minimum = Series(0, index=[metric.min()])

    hist = metric.value_counts(bins=bins, sort=False, normalize=True)
    hist.index = hist.index.right
    hist = concat([minimum, hist])
    hist.index.name, hist.name = metric.name, "frequency"
    return hist.round(6)
